import time
import math
import os

import numpy as np
import jax
import jax.numpy as jnp

from ..utils.atomic_units import AtomicUnits as au
from .thermostats import get_thermostat
from .barostats import get_barostat
from .colvars import setup_colvars
from .spectra import initialize_ir_spectrum

from .utils import load_dynamics_restart, get_restart_file
from .initial import load_model, load_system_data, initialize_preprocessing


def initialize_dynamics(simulation_parameters, fprec, rng_key):
    ### LOAD MODEL
    model = load_model(simulation_parameters)
    model_energy_unit = au.get_multiplier(model.energy_unit)

    ### Get the coordinates and species from the xyz file
    system_data, conformation = load_system_data(simulation_parameters, fprec)

    ### FINISH BUILDING conformation
    if os.path.exists(get_restart_file(system_data)):
        ### RESTART FROM PREVIOUS DYNAMICS
        restart_data = load_dynamics_restart(system_data)
        print("# RESTARTING FROM PREVIOUS DYNAMICS")
        model.preproc_state = restart_data["preproc_state"]
        conformation["coordinates"] = restart_data["coordinates"]
    else:
        restart_data = {}

    ### INITIALIZE PREPROCESSING
    preproc_state, conformation = initialize_preprocessing(
        simulation_parameters, model, conformation, system_data
    )

    ### get dynamics parameters
    dt = simulation_parameters.get("dt") * au.FS
    dt2 = 0.5 * dt
    mass = system_data["mass"]
    totmass_amu = system_data["totmass_amu"]
    nat = system_data["nat"]
    dtm = jnp.asarray(dt / mass[:, None], dtype=fprec)

    nreplicas = system_data.get("nreplicas", 1)
    nbeads = system_data.get("nbeads", None)
    if nbeads is not None:
        nreplicas = nbeads
        dtm = dtm[None, :, :]

    ### INITIALIZE DYNAMICS STATE
    system = {"coordinates": conformation["coordinates"]}
    dyn_state = {
        "istep": 0,
        "dt": dt,
        "pimd": nbeads is not None,
        "preproc_state": preproc_state,
        "start_time_ps": restart_data.get("simulation_time_ps", 0.),
    }
    gradient_keys = ["coordinates"]
    thermo_updates = []

    ### INITIALIZE THERMOSTAT
    thermostat_rng, rng_key = jax.random.split(rng_key)
    (
        thermostat,
        thermostat_post,
        thermostat_state,
        initial_vel,
        dyn_state["thermostat_name"],
    ) = get_thermostat(simulation_parameters, dt, system_data, fprec, thermostat_rng,restart_data)
    do_thermostat_post = thermostat_post is not None
    if do_thermostat_post:
        thermostat_post, post_state = thermostat_post
        dyn_state["thermostat_post_state"] = post_state

    system["thermostat"] = thermostat_state
    system["vel"] = restart_data.get("vel", initial_vel).astype(fprec)

    ### PBC
    pbc_data = system_data.get("pbc", None)
    if pbc_data is not None:
        ### INITIALIZE BAROSTAT
        barostat_key, rng_key = jax.random.split(rng_key)
        thermo_update_ensemble, variable_cell, barostat_state = get_barostat(
            thermostat, simulation_parameters, dt, system_data, fprec, barostat_key,restart_data
        )
        estimate_pressure = variable_cell or pbc_data["estimate_pressure"]
        system["barostat"] = barostat_state
        system["cell"] = conformation["cells"][0]
        if estimate_pressure:
            pressure_o_weight = simulation_parameters.get("pressure_o_weight", 0.0)
            assert (
                0.0 <= pressure_o_weight <= 1.0
            ), "pressure_o_weight must be between 0 and 1"
            gradient_keys.append("strain")
        print("# Estimate pressure: ", estimate_pressure)
    else:
        estimate_pressure = False
        variable_cell = False

        def thermo_update_ensemble(x, v, system):
            v, thermostat_state = thermostat(v, system["thermostat"])
            return x, v, {**system, "thermostat": thermostat_state}

    dyn_state["estimate_pressure"] = estimate_pressure
    dyn_state["variable_cell"] = variable_cell
    thermo_updates.append(thermo_update_ensemble)

    ### ENERGY ENSEMBLE
    ensemble_key = simulation_parameters.get("etot_ensemble_key", None)

    ### COLVARS
    colvars_definitions = simulation_parameters.get("colvars", None)
    use_colvars = colvars_definitions is not None
    if use_colvars:
        colvars_calculators = setup_colvars(colvars_definitions)

    ### IR SPECTRUM
    do_ir_spectrum = simulation_parameters.get("ir_spectrum", False)
    assert isinstance(do_ir_spectrum, bool), "ir_spectrum must be a boolean"
    if do_ir_spectrum:
        is_qtb = dyn_state["thermostat_name"].endswith("QTB")
        model_ir, ir_state, save_dipole, ir_post = initialize_ir_spectrum(
            simulation_parameters, system_data, fprec, dt, is_qtb
        )
        dyn_state["ir_spectrum"] = ir_state

    ### BUILD GRADIENT FUNCTION
    energy_and_gradient = model.get_gradient_function(
        *gradient_keys, jit=True, variables_as_input=True
    )

    ### COLLECT THERMO UPDATES
    if len(thermo_updates) == 1:
        thermo_update = thermo_updates[0]
    else:

        def thermo_update(x, v, system):
            for update in thermo_updates:
                x, v, system = update(x, v, system)
            return x, v, system

    ### RING POLYMER INITIALIZATION
    if nbeads is not None:
        cay_correction = simulation_parameters.get("cay_correction", True)
        omk = system_data["omk"]
        eigmat = system_data["eigmat"]
        cayfact = 1.0 / (4.0 + (dt * omk[1:, None, None]) ** 2) ** 0.5
        if cay_correction:
            axx = jnp.asarray(2 * cayfact)
            axv = jnp.asarray(dt * cayfact)
            avx = jnp.asarray(-dt * cayfact * omk[1:, None, None] ** 2)
        else:
            axx = jnp.asarray(np.cos(omk[1:, None, None] * dt2))
            axv = jnp.asarray(np.sin(omk[1:, None, None] * dt2) / omk[1:, None, None])
            avx = jnp.asarray(-omk[1:, None, None] * np.sin(omk[1:, None, None] * dt2))

        coordinates = conformation["coordinates"].reshape(nbeads, -1, 3)
        eigx = jnp.zeros_like(coordinates).at[0].set(coordinates[0])
        system["coordinates"] = eigx

    ###############################################
    ### DEFINE UPDATE FUNCTION
    @jax.jit
    def update_conformation(conformation, system):
        x = system["coordinates"]
        if nbeads is not None:
            x = jnp.einsum("in,n...->i...", eigmat, x).reshape(nbeads * nat, 3) * (
                nbeads**0.5
            )
        conformation = {**conformation, "coordinates": x}
        if variable_cell:
            conformation["cells"] = system["cell"][None, :, :].repeat(nreplicas, axis=0)

        

        return conformation

    ###############################################
    ### DEFINE INTEGRATION FUNCTIONS
    def integrate_A_half(x0, v0):
        if nbeads is None:
            return x0 + dt2 * v0, v0

        # update coordinates and velocities of a free ring polymer for a half time step
        eigx_c = x0[0] + dt2 * v0[0]
        eigv_c = v0[0]
        eigx = x0[1:] * axx + v0[1:] * axv
        eigv = x0[1:] * avx + v0[1:] * axx

        return (
            jnp.concatenate((eigx_c[None], eigx), axis=0),
            jnp.concatenate((eigv_c[None], eigv), axis=0),
        )

    @jax.jit
    def integrate(system):
        x = system["coordinates"]
        v = system["vel"] + dtm * system["forces"]
        x, v = integrate_A_half(x, v)
        x, v, system = thermo_update(x, v, system)
        x, v = integrate_A_half(x, v)

        return {**system, "coordinates": x, "vel": v}

    ###############################################
    ### DEFINE OBSERVABLE FUNCTION
    @jax.jit
    def update_observables(system, conformation):
        ### POTENTIAL ENERGY AND FORCES
        epot, de, out = energy_and_gradient(model.variables, conformation)
        epot = epot / model_energy_unit
        de = {k: v / model_energy_unit for k, v in de.items()}
        forces = -de["coordinates"]

        if nbeads is not None:
            ### PROJECT FORCES ONTO POLYMER NORMAL MODES
            forces = jnp.einsum(
                "in,i...->n...", eigmat, forces.reshape(nbeads, nat, 3)
            ) * (1.0 / nbeads**0.5)

        system = {
            **system,
            "epot": jnp.mean(epot),
            "forces": forces,
            "energy_gradients": de,
        }

        ### KINETIC ENERGY
        v = system["vel"]
        if nbeads is None:
            corr_kin = system["thermostat"].get("corr_kin", 1.0)
            # ek = 0.5 * jnp.sum(mass[:, None] * v**2) / state_th.get("corr_kin", 1.0)
            ek = (0.5 / nreplicas / corr_kin) * jnp.sum(
                mass[:, None, None] * v[:, :, None] * v[:, None, :], axis=0
            )
        else:
            ek_c = 0.5 * jnp.sum(
                mass[:, None, None] * v[0, :, :, None] * v[0, :, None, :], axis=0
            )
            ek = ek_c - 0.5 * jnp.sum(
                system["coordinates"][1:, :, :, None] * forces[1:, :, None, :],
                axis=(0, 1),
            )
            system["ek_c"] = jnp.trace(ek_c)

        system["ek"] = jnp.trace(ek)
        system["ek_tensor"] = ek

        if estimate_pressure:
            if pressure_o_weight != 1.0:
                v = system["vel"] + 0.5 * dtm * system["forces"]
                if nbeads is None:
                    corr_kin = system["thermostat"].get("corr_kin", 1.0)
                    # ek = 0.5 * jnp.sum(mass[:, None] * v**2) / state_th.get("corr_kin", 1.0)
                    ek = (0.5 / nreplicas / corr_kin) * jnp.sum(
                        mass[:, None, None] * v[:, :, None] * v[:, None, :], axis=0
                    )
                else:
                    ek_c = 0.5 * jnp.sum(
                        mass[:, None, None] * v[0, :, :, None] * v[0, :, None, :],
                        axis=0,
                    )
                    ek = ek_c - 0.5 * jnp.sum(
                        system["coordinates"][1:, :, :, None] * forces[1:, :, None, :],
                        axis=(0, 1),
                    )
                b = pressure_o_weight
                ek = (1.0 - b) * ek + b * system["ek_tensor"]

            vir = jnp.mean(de["strain"], axis=0)
            system["virial"] = vir

            volume = jnp.abs(jnp.linalg.det(system["cell"]))
            Pres = (2 * ek - vir) / volume
            system["pressure_tensor"] = Pres
            system["pressure"] = jnp.trace(Pres) * (1.0 / 3.0)
            if variable_cell:
                density = totmass_amu / volume
                system["density"] = density
                system["volume"] = volume

        if ensemble_key is not None:
            kT = system_data["kT"]
            dE = (
                jnp.mean(out[ensemble_key], axis=0) / model_energy_unit - system["epot"]
            )
            system["ensemble_weights"] = -dE / kT

        if "total_dipole" in out:
            if nbeads is None:
                system["total_dipole"] = out["total_dipole"][0]
            else:
                system["total_dipole"] = jnp.mean(out["total_dipole"], axis=0)

        if use_colvars:
            coords = system["coordinates"][0]
            colvars = {}
            for colvar_name, colvar_calc in colvars_calculators.items():
                colvars[colvar_name] = colvar_calc(coords)
            system["colvars"] = colvars

        return system, out

    ###############################################
    ### IR SPECTRUM
    if do_ir_spectrum:
        # @jax.jit
        # def update_dipole(ir_state,system,conformation):
        #     def mumodel(coords):
        #         out = model_ir._apply(model_ir.variables,{**conformation,"coordinates":coords})
        #         if nbeads is None:
        #             return out["total_dipole"][0]
        #         return out["total_dipole"].sum(axis=0)
        #     dmudqmodel = jax.jacobian(mumodel)

        #     dmudq = dmudqmodel(conformation["coordinates"])
        #     # print(dmudq.shape)
        #     if nbeads is None:
        #         vel = system["vel"].reshape(-1,1,nat,3)[0]
        #         mudot = (vel*dmudq).sum(axis=(1,2))
        #     else:
        #         dmudq = dmudq.reshape(3,nbeads,nat,3)#.mean(axis=1)
        #         vel = (jnp.einsum("in,n...->i...", eigmat, system["vel"]) *  nbeads**0.5
        #         )
        #         # vel = system["vel"][0].reshape(1,nat,3)
        #         mudot = (vel[None,...]*dmudq).sum(axis=(1,2,3))/nbeads

        #     ir_state = save_dipole(mudot,ir_state)
        #     return ir_state
        @jax.jit
        def update_conformation_ir(conformation, system):
            conformation = {
                **conformation,
                "coordinates": system["coordinates"].reshape(-1, nat, 3)[0],
                "natoms": jnp.asarray([nat]),
                "batch_index": jnp.asarray([0] * nat),
                "species": jnp.asarray(system_data["species"].reshape(-1, nat)[0]),
            }
            if variable_cell:
                conformation["cells"] = system["cell"][None, :, :]
                conformation["reciprocal_cells"] = jnp.linalg.inv(system["cell"])[
                    None, :, :
                ]
            return conformation

        @jax.jit
        def update_dipole(ir_state, system, conformation):
            if model_ir is not None:
                out = model_ir._apply(model_ir.variables, conformation)
                q = out.get("charges", jnp.zeros(nat)).reshape((-1, nat))
                dip = out.get("dipoles", jnp.zeros((nat, 3))).reshape((-1, nat, 3))
            else:
                q = system.get("charges", jnp.zeros(nat)).reshape((-1, nat))
                dip = system.get("dipoles", jnp.zeros((nat, 3))).reshape((-1, nat, 3))
            if nbeads is not None:
                q = jnp.mean(q, axis=0)
                dip = jnp.mean(dip, axis=0)
                vel = system["vel"][0]
                pos = system["coordinates"][0]
            else:
                q = q[0]
                dip = dip[0]
                vel = system["vel"].reshape(-1, nat, 3)[0]
                pos = system["coordinates"].reshape(-1, nat, 3)[0]

            if pbc_data is not None:
                cell_reciprocal = (
                    conformation["cells"][0],
                    conformation["reciprocal_cells"][0],
                )
            else:
                cell_reciprocal = None

            ir_state = save_dipole(
                q, vel, pos, dip.sum(axis=0), cell_reciprocal, ir_state
            )
            return ir_state

    ###############################################
    ### GRAPH UPDATES

    nblist_verbose = simulation_parameters.get("nblist_verbose", False)
    nblist_stride = int(simulation_parameters.get("nblist_stride", -1))
    nblist_warmup_time = simulation_parameters.get("nblist_warmup_time", -1.0) * au.FS
    nblist_warmup = int(nblist_warmup_time / dt) if nblist_warmup_time > 0 else 0
    nblist_skin = simulation_parameters.get("nblist_skin", -1.0)
    if nblist_skin > 0:
        if nblist_stride <= 0:
            ## reference skin parameters at 300K (from Tinker-HP)
            ##   => skin of 2 A gives you 40 fs without complete rebuild
            t_ref = 40.0  # FS
            nblist_skin_ref = 2.0  # A
            nblist_stride = int(math.floor(nblist_skin / nblist_skin_ref * t_ref / dt))
        print(
            f"# nblist_skin: {nblist_skin:.2f} A, nblist_stride: {nblist_stride} steps, nblist_warmup: {nblist_warmup} steps"
        )

    if nblist_skin <= 0:
        nblist_stride = 1

    dyn_state["nblist_countdown"] = 0
    dyn_state["print_skin_activation"] = nblist_warmup > 0

    def update_graphs(istep, dyn_state, system, conformation, force_preprocess=False):
        nblist_countdown = dyn_state["nblist_countdown"]
        if nblist_countdown <= 0 or force_preprocess or (istep < nblist_warmup):
            ### FULL NBLIST REBUILD
            dyn_state["nblist_countdown"] = nblist_stride - 1
            preproc_state = dyn_state["preproc_state"]
            conformation = model.preprocessing.process(
                preproc_state, update_conformation(conformation, system)
            )
            preproc_state, state_up, conformation, overflow = (
                model.preprocessing.check_reallocate(preproc_state, conformation)
            )
            dyn_state["preproc_state"] = preproc_state
            if nblist_verbose and overflow:
                print("step", istep, ", nblist overflow => reallocating nblist")
                print("size updates:", state_up)

            if do_ir_spectrum and model_ir is not None:
                conformation_ir = model_ir.preprocessing.process(
                    dyn_state["preproc_state_ir"],
                    update_conformation_ir(dyn_state["conformation_ir"], system),
                )
                (
                    dyn_state["preproc_state_ir"],
                    _,
                    dyn_state["conformation_ir"],
                    overflow,
                ) = model_ir.preprocessing.check_reallocate(
                    dyn_state["preproc_state_ir"], conformation_ir
                )

        else:
            ### SKIN UPDATE
            if dyn_state["print_skin_activation"]:
                if nblist_verbose:
                    print(
                        "step",
                        istep,
                        ", end of nblist warmup phase => activating skin updates",
                    )
                dyn_state["print_skin_activation"] = False

            dyn_state["nblist_countdown"] = nblist_countdown - 1
            conformation = model.preprocessing.update_skin(
                update_conformation(conformation, system)
            )
            if do_ir_spectrum and model_ir is not None:
                dyn_state["conformation_ir"] = model_ir.preprocessing.update_skin(
                    update_conformation_ir(dyn_state["conformation_ir"], system)
                )

        return conformation, dyn_state

    ################################################
    ### DEFINE STEP FUNCTION
    def step(istep, dyn_state, system, conformation, force_preprocess=False):

        dyn_state = {
            **dyn_state,
            "istep": dyn_state["istep"] + 1,
        }

        ### INTEGRATE EQUATIONS OF MOTION
        system = integrate(system)

        ### UPDATE CONFORMATION AND GRAPHS
        conformation, dyn_state = update_graphs(
            istep, dyn_state, system, conformation, force_preprocess
        )

        ## COMPUTE FORCES AND OBSERVABLES
        system, out = update_observables(system, conformation)

        ## END OF STEP UPDATES
        if do_thermostat_post:
            system["thermostat"], dyn_state["thermostat_post_state"] = thermostat_post(
                system["thermostat"], dyn_state["thermostat_post_state"]
            )
        
        if do_ir_spectrum:
            ir_state = update_dipole(
                dyn_state["ir_spectrum"], system, dyn_state["conformation_ir"]
            )
            dyn_state["ir_spectrum"] = ir_post(ir_state)

        return dyn_state, system, conformation, out

    ###########################################################

    print("# Computing initial energy and forces")

    conformation = update_conformation(conformation, system)
    # initialize IR conformation
    if do_ir_spectrum and model_ir is not None:
        dyn_state["preproc_state_ir"], dyn_state["conformation_ir"] = (
            model_ir.preprocessing(
                model_ir.preproc_state,
                update_conformation_ir(conformation, system),
            )
        )

    system, _ = update_observables(system, conformation)

    return step, update_conformation, system_data, dyn_state, conformation, system



import jax
import jax.numpy as jnp
import flax.linen as nn
from ...utils.spherical_harmonics import generate_spherical_harmonics
from ..misc.encodings import SpeciesEncoding, RadialBasis
import dataclasses
import numpy as np
from typing import Any, Dict, List, Union, Callable, Tuple, Sequence, Optional, ClassVar
from ..misc.nets import FullyConnectedNet
from ..misc.e3 import FilteredTensorProduct, ChannelMixingE3, ChannelMixing


class MiniMaceEmbedding(nn.Module):
    """Minimal MACE Embedding

    FID : MINIMACE

    This is a simplified version of the MACE embedding from the paper:
    Batatia et al., MACE: Higher Order Equivariant Message Passing Neural Networks for Fast and Accurate Force Fields
    https://doi.org/10.48550/arXiv.2206.07697

    It is designed to neglect the most costly operations (such as edge-wise tensor products)
    and filter the results at each atomic tensor products to control the number of tensors.
    It may not have the same performance as the full MACE embedding but should be faster.
        
    """
    _graphs_properties: Dict
    dim: int = 128
    """The dimension of the embedding."""
    nchannels: int = 16
    """The number of tensor channels."""
    message_dim: int = 16
    """The dimension of the message formed from the current embedding."""
    nlayers: int = 2
    """The number of interaction layers."""
    ntp: int = 2
    """The number of tensor products per layer."""
    lmax: int = 2
    """The maximum angular momentum of spherical tensors."""
    embedding_hidden: Sequence[int] = dataclasses.field(default_factory=lambda: [])
    """The hidden layers for the species embedding network."""
    latent_hidden: Sequence[int] = dataclasses.field(default_factory=lambda: [128])
    """The hidden layers for the latent update network."""
    activation: Union[Callable, str] = "silu"
    """The activation function."""
    graph_key: str = "graph"
    """The key for the graph input."""
    embedding_key: str = "embedding"
    """The key for the embedding output."""
    tensor_embedding_key: str = "tensor_embedding"
    """The key for the tensor embedding output."""
    species_encoding: dict = dataclasses.field(default_factory=dict)
    """The species encoding parameters. See `fennol.models.misc.encodings.SpeciesEncoding`. """
    radial_basis: dict = dataclasses.field(default_factory=dict)
    """The radial basis parameters. See `fennol.models.misc.encodings.RadialBasis`. """
    ignore_parity: bool = True
    """Whether to ignore parity of irreps in the tensor products"""

    FID: ClassVar[str] = "MINIMACE"


    @nn.compact
    def __call__(self, inputs):
        species = inputs["species"]
        assert (
            len(species.shape) == 1
        ), "Species must be a 1D array (batches must be flattened)"
        # nchannels_density = (
        #     self.nchannels_density
        #     if self.nchannels_density is not None
        #     else self.nchannels
        # )
        # nrep = np.array([2 * l + 1 for l in range(self.lmax + 1)])

        graph = inputs[self.graph_key]
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]
        switch = graph["switch"][:, None]
        cutoff = self._graphs_properties[self.graph_key]["cutoff"]
        radial_basis = (
            RadialBasis(**{**self.radial_basis, "end": cutoff, "name": "RadialBasis"})(
                graph["distances"]
            )
            * switch
        )

        Yij = generate_spherical_harmonics(lmax=self.lmax, normalize=False)(
            graph["vec"] / graph["distances"][:, None]
        )[:, None, :]

        species_encoding = SpeciesEncoding(
            **self.species_encoding, name="SpeciesEncoding"
        )(species)

        xi = FullyConnectedNet(
            neurons=[*self.embedding_hidden, self.dim],
            activation=self.activation,
            use_bias=True,
            name="species_embedding",
        )(species_encoding)

        nchannels_density = self.message_dim * radial_basis.shape[1]

        for layer in range(self.nlayers):
            mi = nn.Dense(
                self.message_dim,
                use_bias=True,
                name=f"species_linear_{layer}",
            )(xi)
            xij = (mi[edge_dst, :, None] * radial_basis[:, None, :]).reshape(
                -1, nchannels_density
            )
            if layer == 0:
                rhoij = xij[:, :, None] * Yij
                density = jax.ops.segment_sum(rhoij, edge_src, species.shape[0])
                Vi = ChannelMixingE3(
                    self.lmax,
                    nchannels_density,
                    self.nchannels,
                    name=f"Vi_initial",
                )(density)
            else:
                rhoi = ChannelMixingE3(
                    self.lmax,
                    self.nchannels,
                    nchannels_density,
                    name=f"rho_mixing_{layer}",
                )(Vi)
                rhoij = xij[:, :, None] * rhoi[edge_dst]
                density = density + jax.ops.segment_sum(
                    rhoij, edge_src, species.shape[0]
                )

            scals = [jax.lax.index_in_dim(density, 0, axis=-1, keepdims=False)]
            for i in range(self.ntp):
                Hi = ChannelMixing(
                    self.lmax,
                    nchannels_density,
                    self.nchannels,
                    name=f"density_mixing_{layer}_{i}",
                )(density)
                Li = FilteredTensorProduct(
                    self.lmax,
                    self.lmax,
                    self.lmax,
                    name=f"TP_{layer}_{i}",
                    ignore_parity=self.ignore_parity,
                )(Vi, Hi)
                scals.append(jax.lax.index_in_dim(Li, 0, axis=-1, keepdims=False))
                Vi = Vi + Li

            dxi = FullyConnectedNet(
                [*self.latent_hidden, self.dim],
                activation=self.activation,
                use_bias=True,
                name=f"latent_net_{layer}",
            )(jnp.concatenate([xi, *scals], axis=-1))
            xi = xi + dxi

        if self.embedding_key is None:
            return xi, Vi
        return {**inputs, self.embedding_key: xi, self.tensor_embedding_key: Vi}



import numpy as np
import jax
import jax.numpy as jnp
import pickle
from ..utils import AtomicUnits as au


def get_restart_file(system_data):
    restart_file = system_data["name"] + ".dyn.restart"
    return restart_file


def load_dynamics_restart(system_data):
    with open(get_restart_file(system_data), "rb") as f:
        restart_data = pickle.load(f)

    assert (
        restart_data["nat"] == system_data["nat"]
    ), f"Restart file does not match system data 'nat'"
    assert restart_data.get("nbeads", 0) == system_data.get(
        "nbeads", 0
    ), f"Restart file does not match system data 'nbeads'"
    assert restart_data.get("nreplicas", 0) == system_data.get(
        "nreplicas", 0
    ), f"Restart file does not match system data 'nreplicas'"
    assert np.all(
        restart_data["species"] == system_data["species"]
    ), f"Restart file does not match system data 'species'"

    return restart_data


def save_dynamics_restart(system_data, conformation, dyn_state, system):
    restart_data = {
        "nat": system_data["nat"],
        "nbeads": system_data.get("nbeads", 0),
        "nreplicas": system_data.get("nreplicas", 0),
        "species": system_data["species"],
        "coordinates": conformation["coordinates"],
        "vel": system["vel"],
        "preproc_state": dyn_state["preproc_state"],
        "simulation_time_ps": dyn_state["start_time_ps"] + (dyn_state["dt"]*1e-3)* dyn_state["istep"],
    }
    if "cells" in conformation:
        restart_data["cells"] = conformation["cells"]

    

    restart_file = get_restart_file(system_data)
    with open(restart_file, "wb") as f:
        pickle.dump(restart_data, f)


@jax.jit
def wrapbox(x, cell, reciprocal_cell):
    # q = jnp.einsum("ji,sj->si", reciprocal_cell, x)
    q = x @ reciprocal_cell
    q = q - jnp.floor(q)
    # return jnp.einsum("ji,sj->si", cell, q)
    return q @ cell


def test_pressure_fd(system_data, conformation, model, verbose=True):
    model_energy_unit = au.get_multiplier(model.energy_unit)
    nat = system_data["nat"]
    volume = system_data["pbc"]["volume"]
    coordinates = conformation["coordinates"]
    cell = conformation["cells"][0]
    # temper = 2 * ek / (3.0 * nat) * au.KELVIN
    ek = 1.5 * nat * system_data["kT"]
    Pkin = (2 * au.KBAR) * ek / ((3.0 / au.BOHR**3) * volume)
    e, f, vir_t, _ = model._energy_and_forces_and_virial(model.variables, conformation)
    KBAR = au.KBAR / model_energy_unit
    Pvir = -(np.trace(vir_t[0]) * KBAR) / ((3.0 / au.BOHR**3) * volume)
    vstep = volume * 0.000001
    scalep = ((volume + vstep) / volume) ** (1.0 / 3.0)
    cellp = cell * scalep
    reciprocal_cell = np.linalg.inv(cellp)
    sysp = model.preprocess(
        **{
            **conformation,
            "coordinates": coordinates * scalep,
            "cells": cellp[None, :, :],
            "reciprocal_cells": reciprocal_cell[None, :, :],
        }
    )
    ep, _ = model._total_energy(model.variables, sysp)
    scalem = ((volume - vstep) / volume) ** (1.0 / 3.0)
    cellm = cell * scalem
    reciprocal_cell = np.linalg.inv(cellm)
    sysm = model.preprocess(
        **{
            **conformation,
            "coordinates": coordinates * scalem,
            "cells": cellm[None, :, :],
            "reciprocal_cells": reciprocal_cell[None, :, :],
        }
    )
    em, _ = model._total_energy(model.variables, sysm)
    Pvir_fd = -(ep[0] * KBAR - em[0] * KBAR) / (2.0 * vstep / au.BOHR**3)
    if verbose:
        print(
            f"# Initial pressure: {Pkin+Pvir:.3f} (virial); {Pkin+Pvir_fd:.3f} (finite difference) ; Pkin: {Pkin:.3f} ; Pvir: {Pvir:.3f} ; Pvir_fd: {Pvir_fd:.3f}"
        )
    return Pkin, Pvir, Pvir_fd



#!/usr/bin/env python3
import sys
import time
import numpy as np


def xyz_reader(filename,box_info=False,indexed=False
                  ,start=1, stop=-1,step=1,max_frames=None
                  ,stream=False,interval=2.,sleep=time.sleep):
  if stop>0 and start > stop: return
  pf=open(filename,"r")
  inside_frame=False
  nat_read=False
  iframe=0
  stride=step
  nframes=0
  if start>1 : print(f"skipping {start-1} frames")
  if not box_info: box=None
  while True:
    line=pf.readline()

    if not line:
      if stream:
        sleep(interval)
        continue
      elif inside_frame or nat_read:
        raise Exception("Error: premature end of file!")
      else:
        break
    
    line = line.strip()
    if line.startswith("#"): continue

    if not inside_frame:
      if not nat_read:
        nat=int(line.split()[0])
        nat_read=True
        iat=0
        inside_frame = not box_info
        xyz=np.zeros((nat,3))
        species=[]
        continue
      
      box=np.array(line.split(),dtype="float")
      inside_frame=True
      continue

    ls=line.split()
    iat,s = (int(ls[0]),1) if indexed else (iat+1,0)
    species.append(ls[s])
    xyz[iat-1,:] = np.array([ls[s+1],ls[s+2],ls[s+3]],dtype="float")
    if iat == nat:
      iframe+=1
      inside_frame=False
      nat_read=False
      if iframe < start: continue
      stride+=1
      if stride>=step:
        nframes+=1
        stride=0
        yield np.array(species,dtype='<U2'),xyz,box
      if (max_frames is not None and nframes >= max_frames) \
          or (stop>0 and iframe >= stop):
        break

def apply_pbc(vec,cell):
  cellinv = np.linalg.inv(cell)
  #q = cellinv @ vec.T
  q=np.einsum("ij,sj->si",cellinv,vec)
  q-=np.round(q)
  #return (cell @ q).T
  return np.einsum("ij,sj->si",cell,q)

def get_pairs(species,sp1=None,sp2=None):
  #nat=len(species)
  #id1=np.flatnonzero(species==sp1) if sp1 is not None else np.arange(nat)
  #if sp2 != sp1:
  #  id2=np.flatnonzero(species==sp2) if sp2 is not None else np.arange(nat)
  #pairs=np.array([np.repeat(id1, len(id2)),np.tile(id2, len(id1))])
  #if sp1 is None or sp2 is None or sp1==sp2:
  #  pairs=pairs[:,pairs[0]!=pairs[1]]
  #return pairs
  #if sp1 is None and sp2 is not None: sp1=sp2
  #if sp2 is None and sp1 is not None: sp2=sp1
  if sp1 is None:
    if sp2 is None: 
      return np.array(np.triu_indices(len(species),1)),0.5*len(species)**2
    sp1=sp2
  id1=np.flatnonzero(species==sp1)
  num1=len(id1)
  if sp2==sp1 or sp2 is None:
    pairsloc=np.triu_indices(len(id1),1)
    return np.array([id1[pairsloc[0]],id1[pairsloc[1]]]),0.5*num1**2
  id2=np.flatnonzero(species==sp2)
  num2=len(id2)
  return np.array([np.repeat(id1, len(id2)),np.tile(id2, len(id1))]),num1*num2

def radial_density(species,xyz,cell=None,rmax=10.0,dr=0.1,sp1=None,sp2=None,pairs=None,timer=False):
  nat = len(species)
  if timer: 
    print(sp1,sp2)
    time0=time.time()
  if pairs is None:
    pairs,n_pairs_norm=get_pairs(species,sp1,sp2)
  n_pairs=pairs.shape[1]
  mult = 1 if sp1!=sp2 else 2
  #print(mult*n_pairs,mult)
  if timer:
    print("pairs:",time.time()-time0)
    time0=time.time()
  vec=xyz[pairs[0]]-xyz[pairs[1]]
  if cell is not None:
    vec=apply_pbc(vec,cell)
  dist=np.linalg.norm(vec,axis=1)
  if timer: 
    print("dists:",time.time()-time0)
    time0=time.time()
  n_bins=int(rmax/dr)
  bins=np.linspace(0,rmax,n_bins+1)
  r,_=np.histogram(dist,bins=n_bins,range=(0,rmax))
  if timer: print("hist:",time.time()-time0)
  expect = n_pairs_norm*4./3.*np.pi*((bins[1:])**3 - bins[:-1]**3)
  if cell is not None:
    volume=np.dot(cell[:,0],np.cross(cell[:,1],cell[:,2]))
    expect/=volume
  return r/expect


def radial_from_file(xyzfile,pairs,outfile="gr.dat",rmax=10,dr=0.02,thermalize=0,stride=1,box_info=False,indexed=False,watch=False,cell=None,**kwargs):
  if (('tinker' in kwargs and kwargs["tinker"])):
    indexed=True
    box_info=True
  
  try:
    #frames = list(read_arc_file(arcfile,max_frames=205))
    #print(len(frames))
    reader=xyz_reader(xyzfile,box_info=box_info,indexed=indexed
                ,start=thermalize+1,step=stride,stream=watch)
    n_bins=int(rmax/dr)
    centers=(np.arange(n_bins)+0.5)*dr 
    nframe=0
    if cell is not None:
      #cell=np.array([float(c) for c in cell.split()]).reshape((3,3),order="F")
      cell = np.array(cell).reshape((3,3),order="F")
      print("cell matrix=")
      print(cell)
    #setup_figures(["g(r)"])
    #plt.pause(1.)
    pairs = [p.strip().replace("-"," ").replace("_"," ") for p in pairs]
    header = " ".join(["r"]+["-".join(p.split()) for p in pairs])
    box=None
    gr_avg=np.zeros((n_bins,len(pairs)))
    for frame in reader:
      species,xyz,box=frame
      if box is not None:
        cell=np.diag(box[:3])
      nframe+=1
      print(nframe)
      for i,pair in enumerate(pairs):
        sp1,sp2=pair.split()
        gr=radial_density(species,xyz,cell,rmax,dr,sp1=sp1,sp2=sp2,timer=False)
        #gr=radial_density_numba(species,xyz,box,rmax,dr,sp1=sp1,sp2=sp2)
        gr_avg[:,i]+=gr
      #cleanup_ax(plt.gca())
      #plt.plot(centers,gr_avg/nframe)
      #plt.pause(1.)
      np.savetxt(outfile,np.column_stack((centers,gr_avg/nframe)),header=header)
      
  except KeyboardInterrupt:
    sys.exit(0)

def main(xyzfile:str,thermalize:int=0,watch:bool=False):

  # box length
  a=18.643
  cell=np.array([a,0,0,0,a,0,0,0,a]).reshape((3,3),order="F")

  # maximum distance to compute the rdf
  rmax=9.
  # bin width
  dr=0.05

  # pairs of atoms to compute the rdf
  pairs=["O O","O H","H H"]

  radial_from_file(xyzfile,pairs,cell=cell,rmax=rmax,dr=dr,indexed=True,box_info=False
                  ,watch=watch,thermalize=thermalize)

if __name__=="__main__":
    import argparse
    parser = argparse.ArgumentParser(description='compute radial distribution function')
    parser.add_argument('arcfile', type=str, help='arc file')
    parser.add_argument('--thermalize','-t', type=int, default=0, help='number of frames to skip')
    parser.add_argument('--watch','-w', action='store_true', help='watch file')
    args = parser.parse_args()
    main(args.arcfile,args.thermalize,args.watch)
    





import numpy as np
import jax
import math
import jax.numpy as jnp
import sympy
from sympy.printing.pycode import pycode
from sympy.physics.wigner import clebsch_gordan
from functools import partial


def CG_SU2(j1: int, j2: int, j3: int) -> np.array:
    r"""Clebsch-Gordan coefficients for the direct product of two irreducible representations of :math:`SU(2)`
    Returns
    -------
    `np.Array`
        tensor :math:`C` of shape :math:`(2j_1+1, 2j_2+1, 2j_3+1)`
    """
    C = np.zeros((2 * j1 + 1, 2 * j2 + 1, 2 * j3 + 1))
    for m1 in range(-j1, j1 + 1):
        for m2 in range(-j2, j2 + 1):
            for m3 in range(-j3, j3 + 1):
                C[m1 + j1, m2 + j2, m3 + j3] = float(
                    clebsch_gordan(j1, j2, j3, m1, m2, m3)
                )
    return C


def change_basis_real_to_complex(l: int) -> np.array:
    r"""Change of basis matrix from real to complex spherical harmonics
    https://en.wikipedia.org/wiki/Spherical_harmonics#Real_form
    adapted from e3nn.o3._wigner
    """
    q = np.zeros((2 * l + 1, 2 * l + 1), dtype=complex)
    for m in range(-l, 0):
        q[l + m, l + abs(m)] = 1 / 2**0.5
        q[l + m, l - abs(m)] = -1j / 2**0.5
    q[l, l] = 1
    for m in range(1, l + 1):
        q[l + m, l + abs(m)] = (-1) ** m / 2**0.5
        q[l + m, l - abs(m)] = 1j * (-1) ** m / 2**0.5

    # factor of (-i)**l to make the Clebsch-Gordan coefficients real
    return q * (-1j) ** l


def CG_SO3(j1: int, j2: int, j3: int) -> np.array:
    r"""Clebsch-Gordan coefficients for the direct product of two irreducible representations of :math:`SO(3)`
    Returns
    -------
    `np.array`
        tensor :math:`C` of shape :math:`(2l_1+1, 2l_2+1, 2l_3+1)`
    """
    C = CG_SU2(j1, j2, j3)
    Q1 = change_basis_real_to_complex(j1)
    Q2 = change_basis_real_to_complex(j2)
    Q3 = change_basis_real_to_complex(j3)
    C = np.real(np.einsum("ij,kl,mn,ikn->jlm", Q1, Q2, np.conj(Q3.T), C))
    return C / np.linalg.norm(C)


def generate_spherical_harmonics(
    lmax, normalize=False, print_code=False, jit=False, vmapped=False
):  # pragma: no cover
    r"""returns a function that computes spherical harmonic up to lmax
    (adapted from e3nn)
    """

    def to_frac(x: float):
        from fractions import Fraction

        s = 1 if x >= 0 else -1
        x = x**2
        x = Fraction(x).limit_denominator()
        x = s * sympy.sqrt(x)
        x = sympy.simplify(x)
        return x

    if vmapped:
        fn_str = "def spherical_harmonics_(x,y,z):\n"
        fn_str += "  sh_0_0 = 1.\n"
    else:
        fn_str = "def spherical_harmonics_(vec):\n"
        if normalize:
            fn_str += "  vec = vec/jnp.linalg.norm(vec,axis=-1,keepdims=True)\n"
        fn_str += "  x,y,z = [jax.lax.index_in_dim(vec, i, axis=-1, keepdims=False) for i in range(3)]\n"
        fn_str += "  sh_0_0 = jnp.ones_like(x)\n"

    x_var, y_var, z_var = sympy.symbols("x y z")
    polynomials = [sympy.sqrt(3) * x_var, sympy.sqrt(3) * y_var, sympy.sqrt(3) * z_var]

    def sub_z1(p, names, polynormz):
        p = p.subs(x_var, 0).subs(y_var, 1).subs(z_var, 0)
        for n, c in zip(names, polynormz):
            p = p.subs(n, c)
        return p

    poly_evalz = [sub_z1(p, [], []) for p in polynomials]

    for l in range(1, lmax + 1):
        sh_variables = sympy.symbols(" ".join(f"sh_{l}_{m}" for m in range(2 * l + 1)))

        for n, p in zip(sh_variables, polynomials):
            fn_str += f"  {n} = {pycode(p.evalf())}\n"

        if l == lmax:
            break

        polynomials = [
            sum(
                to_frac(c.item()) * v * sh
                for cj, v in zip(cij, [x_var, y_var, z_var])
                for c, sh in zip(cj, sh_variables)
            )
            for cij in CG_SO3(l + 1, 1, l)
        ]

        poly_evalz = [sub_z1(p, sh_variables, poly_evalz) for p in polynomials]
        norm = sympy.sqrt(sum(p**2 for p in poly_evalz))
        polynomials = [sympy.sqrt(2 * l + 3) * p / norm for p in polynomials]
        poly_evalz = [sympy.sqrt(2 * l + 3) * p / norm for p in poly_evalz]

        polynomials = [sympy.simplify(p, full=True) for p in polynomials]

    u = ",\n        ".join(
        ", ".join(f"sh_{j}_{m}" for m in range(2 * j + 1)) for j in range(l + 1)
    )
    if vmapped:
        fn_str += f"  return jnp.array([\n        {u}\n    ])\n"
    else:
        fn_str += f"  return jnp.stack([\n        {u}\n    ], axis=-1)\n"

    if print_code:
        print(fn_str)
    exec(fn_str)
    sh = locals()["spherical_harmonics_"]
    if jit:
        sh = jax.jit(sh)
    if not vmapped:
        return sh

    if normalize:

        def spherical_harmonics(vec):
            vec = vec / jnp.linalg.norm(vec, axis=-1, keepdims=True)
            x, y, z = [
                jax.lax.index_in_dim(vec, i, axis=-1, keepdims=False) for i in range(3)
            ]
            return jax.vmap(sh)(x, y, z)

    else:

        def spherical_harmonics(vec):
            x, y, z = [
                jax.lax.index_in_dim(vec, i, axis=-1, keepdims=False) for i in range(3)
            ]
            return jax.vmap(sh)(x, y, z)

    if jit:
        spherical_harmonics = jax.jit(spherical_harmonics)
    return spherical_harmonics


@partial(jax.jit, static_argnums=1)
def spherical_to_cartesian_tensor(Q, lmax):
    q = Q[..., 0]
    if lmax == 0:
        return q[..., None]

    mu = Q[..., 1:4]
    if lmax == 1:
        return jnp.concatenate([q[..., None], mu], axis=-1)

    Q22s = Q[..., 4]
    Q21s = Q[..., 5]
    Q20 = Q[..., 6]
    Q21c = Q[..., 7]
    Q22c = Q[..., 8]
    Tzz = -0.5 * Q20 + (0.5 * 3**0.5) * Q22c
    Txx = -0.5 * Q20 - (0.5 * 3**0.5) * Q22c
    Tyy = Q20
    Txz = 0.5 * (3**0.5) * Q22s
    Tyz = 0.5 * (3**0.5) * Q21c
    Txy = 0.5 * (3**0.5) * Q21s

    if lmax == 2:
        return jnp.concatenate(
            [
                q[..., None],
                mu,
                Txx[..., None],
                Tyy[..., None],
                Tzz[..., None],
                Txy[..., None],
                Txz[..., None],
                Tyz[..., None],
            ],
            axis=-1,
        )



import jax
import jax.numpy as jnp
import flax.linen as nn
from typing import Sequence, Dict, Optional, Union, Callable, ClassVar
import numpy as np
import dataclasses

from ..misc.nets import FullyConnectedNet, ChemicalNet

from ...utils.periodic_table import PERIODIC_TABLE, VALENCE_STRUCTURE
from ..misc.encodings import SpeciesEncoding, RadialBasis


class DeepPotEmbedding(nn.Module):
    """Deep Potential embedding

    FID : DEEPPOT

    ### Reference
    Zhang, L., Han, J., Wang, H., Car, R., & E, W. (2018). Deep Potential Molecular dynamics: A scalable model with the accuracy of quantum mechanics. Phys. Rev. Lett., 120(14), 143001. https://doi.org/10.1103/PhysRevLett.120.143001

    """
    _graphs_properties: Dict
    dim: int = 64
    """The dimension of the embedding."""
    subdim: int = 8
    """The first dimensions to select for the embedding tensor product."""
    radial_dim: Optional[int] = None
    """The dimension of the radial embedding for tensor combination. 
        If None, we use a neural net to combine chemical and radial information, like in the original DeepPot."""
    embedding_key: str = "embedding"
    """The key to use for the output embedding in the returned dictionary."""
    graph_key: str = "graph"
    """The key in the input dictionary that corresponds to the radial graph."""
    species_encoding: dict = dataclasses.field(default_factory=dict)
    """The species encoding parameters. See `fennol.models.misc.encodings.SpeciesEncoding`"""
    radial_basis: Optional[dict] = None
    """The radial basis parameters. See `fennol.models.misc.encodings.RadialBasis`. 
        If None, the radial basis is the s_ij like in the original DeepPot."""
    embedding_hidden: Sequence[int] = dataclasses.field(
        default_factory=lambda: [64, 64, 64]
    )
    """The hidden layers of the embedding network."""
    activation: Union[Callable, str] = "silu"
    """The activation function."""
    concatenate_species: bool = False
    """Whether to concatenate the species encoding with the embedding."""
    divide_distances: bool = True
    """Whether to divide the switch by the distance in s_ij."""
    species_order: Optional[Union[str,Sequence[str]]] = None
    """Species considered by the network when using species-specialized embedding network."""

    FID: ClassVar[str] = "DEEPPOT"

    @nn.compact
    def __call__(self, inputs):
        species = inputs["species"]

        # species encoding
        if self.species_order is None or self.concatenate_species:
            onehot = SpeciesEncoding(**self.species_encoding, name="SpeciesEncoding")(
                species
            )

        # Radial graph
        graph = inputs[self.graph_key]
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]
        distances = graph["distances"][:, None]
        switch = graph["switch"][:, None]
        vec = graph["vec"] / distances
        sij = switch / distances if self.divide_distances else switch
        Rij = jnp.concatenate((sij, sij * vec), axis=-1)

        # Radial BASIS
        cutoff = self._graphs_properties[self.graph_key]["cutoff"]
        if self.radial_basis is not None:
            radial_terms = RadialBasis(
                **{
                    **self.radial_basis,
                    "end": cutoff,
                    "name": f"RadialBasis",
                }
            )(graph["distances"])
        else:
            radial_terms = sij

        if self.species_order is not None:
            Gij = ChemicalNet(
                self.species_order,
                [*self.embedding_hidden, self.dim],
                activation=self.activation,
            )((species[edge_dst], radial_terms))
        elif self.radial_dim is not None:
            Gij = FullyConnectedNet(
                [*self.embedding_hidden, self.radial_dim], activation=self.activation
            )(radial_terms)
            Wa = self.param(
                f"Wa",
                nn.initializers.normal(
                    stddev=1.0 / (Gij.shape[1] * onehot.shape[1]) ** 0.5
                ),
                (onehot.shape[1], Gij.shape[1], self.dim),
            )
            Gij = jnp.einsum(
                "...i,...j,ijk->...k",
                onehot[edge_dst],
                Gij,
                Wa,
            )
        else:
            Gij = FullyConnectedNet(
                [*self.embedding_hidden, self.dim], activation=self.activation
            )(jnp.concatenate((radial_terms, onehot[edge_dst]), axis=-1))

        GRi = jax.ops.segment_sum(
            Gij[:, None, :] * Rij[:, :, None], edge_src, species.shape[0]
        )
        if self.subdim > 0:
            GRisub = GRi[:, :, : self.subdim]

            embedding = (
                (GRi[:, :, :, None] * GRisub[:, :, None, :])
                .sum(axis=1)
                .reshape((species.shape[0], -1))
            )
        else:
            GRisub = nn.Dense(self.dim, use_bias=False, name="Gri_linear")(GRi)
            embedding = (GRi * GRisub).sum(axis=1)

        if self.concatenate_species:
            embedding = jnp.concatenate((onehot, embedding), axis=-1)

        if self.embedding_key is None:
            return embedding
        return {**inputs, self.embedding_key: embedding}


class DeepPotE3Embedding(nn.Module):
    """Deep Potential embedding with angle information

    FID : DEEPPOT_E3

    ### Reference
    L. Zhang, J. Han, H. Wang, W. A. Saidi, R. Car, Weinan E, End-to-end Symmetry Preserving Inter-atomic Potential Energy Model for Finite and Extended Systems,
    Conference on Neural Information Processing Systems (NeurIPS), 2018,
    https://doi.org/10.48550/arXiv.1805.09003

    """
    _graphs_properties: Dict
    dim: int = 64
    """The dimension of the embedding."""
    embedding_key: str = "embedding"
    """The key to use for the output embedding in the returned dictionary."""
    graph_key: str = "graph"
    """The key in the input dictionary that corresponds to the graph."""
    species_encoding: dict = dataclasses.field(default_factory=dict)
    """The species encoding parameters. See `fennol.models.misc.encodings.SpeciesEncoding`"""
    embedding_hidden: Sequence[int] = dataclasses.field(
        default_factory=lambda: [64, 64, 64]
    )
    """The hidden layers of the embedding network."""
    activation: Union[Callable, str] = "silu"
    """The activation function."""
    concatenate_species: bool = False
    """Whether to concatenate the species encoding with the embedding."""
    divide_distances: bool = True
    """Whether to divide the switch by the distance in s_ij."""

    FID: ClassVar[str] = "DEEPPOT_E3"

    @nn.compact
    def __call__(self, inputs):
        species = inputs["species"]

        # species encoding
        onehot = SpeciesEncoding(**self.species_encoding, name="SpeciesEncoding")(
            species
        )

        # Radial graph
        graph = inputs[self.graph_key]
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]
        distances = graph["distances"][:, None]
        switch = graph["switch"][:, None]
        vec = graph["vec"] / distances
        sij = switch / distances if self.divide_distances else switch
        Rij = jnp.concatenate((sij, sij * vec), axis=-1)

        zdest = onehot[edge_dst]

        angle_src, angle_dst = graph["angle_src"], graph["angle_dst"]
        z_angsrc = zdest[angle_src]
        z_angdst = zdest[angle_dst]

        # Radial BASIS
        assert (
            "angles" in graph
        ), "Error: DeepPotE3 requires angles (GRAPH_ANGLE_EXTENSION)"
        theta = (Rij[angle_src] * Rij[angle_dst]).sum(axis=-1,keepdims=True)

        Ne3 = FullyConnectedNet(
            [*self.embedding_hidden, self.dim],
            activation=self.activation,
            name="Ne3",
        )
        Gijk = Ne3(jnp.concatenate((theta, z_angsrc, z_angdst), axis=-1)) + Ne3(
            jnp.concatenate((theta, z_angdst, z_angsrc), axis=-1)
        )

        embedding = jax.ops.segment_sum(
            Gijk * theta, graph["central_atom"], species.shape[0]
        )

        if self.concatenate_species:
            embedding = jnp.concatenate((onehot, embedding), axis=-1)

        if self.embedding_key is None:
            return embedding
        return {**inputs, self.embedding_key: embedding}



import numpy as np
import flax.linen as nn
import jax
import jax.numpy as jnp
import math
import optax
from enum import Enum

from ..utils.atomic_units import AtomicUnits as au  # CM1,THZ,BOHR,MPROT
from ..utils import Counter
from ..utils.deconvolution import (
    deconvolute_spectrum,
    kernel_lorentz_pot,
    kernel_lorentz,
)


def get_barostat(
    thermostat, simulation_parameters, dt, system_data, fprec, rng_key=None, restart_data={}
):
    state = {}

    barostat_name = str(simulation_parameters.get("barostat", "NONE")).upper()

    kT = system_data.get("kT", None)
    assert kT is not None, "kT must be specified for NPT/NPH simulations"
    target_pressure = simulation_parameters.get("target_pressure")
    if barostat_name != "NONE":
        assert (
            target_pressure is not None
        ), "target_pressure must be specified for NPT/NPH simulations"
        target_pressure = target_pressure / au.BOHR**3

    nbeads = system_data.get("nbeads", None)
    variable_cell = True

    anisotropic = simulation_parameters.get("aniso_barostat", False)
    
    isotropic = not anisotropic

    pbc_data = system_data["pbc"]
    start_barostat = simulation_parameters.get("start_barostat", 0.0)*au.FS
    start_time = restart_data.get("simulation_time_ps",0.) * 1e3
    start_barostat = max(0.,start_barostat-start_time)
    istart_barostat = int(round(start_barostat / dt))
    if istart_barostat > 0 and barostat_name not in ["NONE"]:
        print(
            f"# BAROSTAT will start at {start_barostat/1000:.3f} ps ({istart_barostat} steps)"
        )
    else:
        istart_barostat = 0

    if barostat_name in ["LGV", "LANGEVIN"]:
        assert rng_key is not None, "rng_key must be provided for QTB barostat"
        gamma = simulation_parameters.get("gamma_piston", 20.0 / au.THZ) / au.FS
        tau_piston = simulation_parameters.get("tau_piston", 200.0 / au.FS) * au.FS
        nat = system_data["nat"]
        masspiston = 3 * nat * kT * tau_piston**2
        print(f"# LANGEVIN barostat with piston mass={masspiston:.1e} Ha.fs^2")
        a1 = math.exp(-gamma * dt)
        a2 = ((1 - a1 * a1) * kT / masspiston) ** 0.5


        rng_key, v_key = jax.random.split(rng_key)
        if anisotropic:
            extvol = pbc_data["cell"]
            vextvol = (
                jax.random.normal(v_key, (3, 3), dtype=extvol.dtype)
                * (kT / masspiston) ** 0.5
            )
            vextvol = 0.5 * (vextvol + vextvol.T)

            aniso_mask = simulation_parameters.get(
                "aniso_mask", [True, True, True, True, True, True]
            )
            assert len(aniso_mask) == 6, "aniso_mask must have 6 elements"
            aniso_mask = np.array(aniso_mask, dtype=bool).astype(np.int32)
            ndof_piston = np.sum(aniso_mask)
            # xx   yy   zz    xy   xz   yz
            aniso_mask = np.array(
                [
                    [aniso_mask[0], aniso_mask[3], aniso_mask[4]],
                    [aniso_mask[3], aniso_mask[1], aniso_mask[5]],
                    [aniso_mask[4], aniso_mask[5], aniso_mask[2]],
                ]
            ,dtype=np.int32)
        else:
            extvol = jnp.asarray(pbc_data["volume"])
            vextvol = (
                jax.random.normal(v_key, (1,), dtype=extvol.dtype)
                * (kT / masspiston) ** 0.5
            )
            ndof_piston = 1.

        state["extvol"] = extvol
        state["vextvol"] = vextvol
        state["rng_key"] = rng_key
        state["istep"] = 0

        def barostat(x, vel, system):
            if nbeads is not None:
                x, eigx = x[0], x[1:]
                vel, eigv = vel[0], vel[1:]
            barostat_state = system["barostat"]
            extvol = barostat_state["extvol"]
            vextvol = barostat_state["vextvol"]
            cell = system["cell"]
            volume = jnp.abs(jnp.linalg.det(cell))

            istep = barostat_state["istep"] + 1
            dt_bar = dt * (istep >= istart_barostat)


            # apply B
            pV = 2 * (system["ek_tensor"] + jnp.trace(system["ek_tensor"])*jnp.eye(3)/(3*x.shape[0])) - system["virial"]
            if isotropic:
                dpV = jnp.trace(pV) - 3*volume * target_pressure
            else:
                dpV = 0.5 * (pV + pV.T) - volume * target_pressure * jnp.eye(3) 

            vextvol = vextvol + (dt_bar/masspiston) * dpV

            # apply A
            if isotropic:
                scale1 = jnp.exp((0.5 * dt_bar*(1+1./x.shape[0])) * vextvol)
                vel = vel / scale1
            else:
                vextvol = aniso_mask * vextvol
                l, O = jnp.linalg.eigh(vextvol + jnp.trace(vextvol) * jnp.eye(3)/(3*x.shape[0]))
                Dv = jnp.diag(jnp.exp(-0.5 * dt_bar * l))
                Dx = jnp.diag(jnp.exp(0.5 * dt_bar * l))
                scalev = O @ Dv @ O.T
                scale1 = O @ Dx @ O.T
                vel = vel @ scalev

            # apply O
            if nbeads is not None:
                eigv, thermostat_state = thermostat(
                    jnp.concatenate((vel[None], eigv), axis=0), system["thermostat"]
                )
                vel, eigv = eigv[0], eigv[1:]
            else:
                vel, thermostat_state = thermostat(vel, system["thermostat"])
            rng_key, noise_key = jax.random.split(barostat_state["rng_key"])

            if isotropic:
                noise = jax.random.normal(noise_key, (1,), dtype=vextvol.dtype)
            else:
                noise = jax.random.normal(noise_key, (3, 3), dtype=vextvol.dtype)
                noise = 0.5 * (noise + noise.T)

            vextvol = a1 * vextvol + a2 * noise

            # apply A
            if isotropic:
                scale2 = jnp.exp((0.5 * dt_bar*(1+1./x.shape[0])) * vextvol)
                # vel = vel * scale1
                vel = vel / scale2
                x = x * (scale1 * scale2)
                extvol = extvol * (scale1 * scale2) ** 3
                cell = cell * (scale1 * scale2)
            else:
                vextvol = aniso_mask * vextvol
                l, O = jnp.linalg.eigh(vextvol + jnp.trace(vextvol) * jnp.eye(3)/(3*x.shape[0]))
                Dv = jnp.diag(jnp.exp(-0.5 * dt_bar * l))
                Dx = jnp.diag(jnp.exp(0.5 * dt_bar * l))
                scalev = O @ Dv @ O.T
                scale = scale1 @ (O @ Dx @ O.T)
                vel = vel @ scalev
                x = x @ scale
                extvol = extvol @ scale
                cell = extvol

            if nbeads is not None:
                x = jnp.concatenate((x[None], eigx), axis=0)
                vel = jnp.concatenate((vel[None], eigv), axis=0)

            piston_temperature = (au.KELVIN * masspiston/ndof_piston) * jnp.sum(vextvol**2)
            barostat_state = {
                **barostat_state,
                "istep": istep,
                "rng_key": rng_key,
                "vextvol": vextvol,
                "extvol": extvol,
                "piston_temperature": piston_temperature,
            }
            return (
                x,
                vel,
                {
                    **system,
                    "barostat": barostat_state,
                    "cell": cell,
                    "thermostat": thermostat_state,
                },
            )

    elif barostat_name in ["NONE"]:
        variable_cell = False

        def barostat(x, vel, system):
            vel, thermostat_state = thermostat(vel, system["thermostat"])
            return x, vel, {**system, "thermostat": thermostat_state}

    else:
        raise ValueError(f"Unknown barostat {barostat_name}")

    return barostat, variable_cell, state



import os, io, sys
import numpy as np
from scipy.spatial.transform import Rotation
from collections import defaultdict
import pickle
import glob
from flax import traverse_util
from typing import Dict, List, Tuple, Union, Optional, Callable, Sequence
from .databases import DBDataset, H5Dataset
from ..models.preprocessing import AtomPadding
import re
import json
import yaml

try:
    import tomlkit
except ImportError:
    tomlkit = None

try:
    from torch.utils.data import DataLoader
except ImportError:
    raise ImportError(
        "PyTorch is required for training models. Install the CPU version following instructions at https://pytorch.org/get-started/locally/"
    )

from ..models import FENNIX


def load_configuration(config_file: str) -> Dict[str, any]:
    if config_file.endswith(".json"):
        parameters = json.load(open(config_file))
    elif config_file.endswith(".yaml") or config_file.endswith(".yml"):
        parameters = yaml.load(open(config_file), Loader=yaml.FullLoader)
    elif tomlkit is not None and config_file.endswith(".toml"):
        parameters = tomlkit.loads(open(config_file).read())
    else:
        supported_formats = [".json", ".yaml", ".yml"]
        if tomlkit is not None:
            supported_formats.append(".toml")
        raise ValueError(
            f"Unknown config file format. Supported formats: {supported_formats}"
        )
    return parameters


def load_dataset(
    dspath: str,
    batch_size: int,
    batch_size_val: Optional[int] = None,
    rename_refs: dict = {},
    infinite_iterator: bool = False,
    atom_padding: bool = False,
    ref_keys: Optional[Sequence[str]] = None,
    split_data_inputs: bool = False,
    np_rng: Optional[np.random.Generator] = None,
    train_val_split: bool = True,
    training_parameters: dict = {},
    add_flags: Sequence[str] = ["training"],
    fprec: str = "float32",
):
    """
    Load a dataset from a pickle file and return two iterators for training and validation batches.

    Args:
        training_parameters (dict): A dictionary with the following keys:
            - 'dspath': str. Path to the pickle file containing the dataset.
            - 'batch_size': int. Number of samples per batch.
        rename_refs (list, optional): A list of strings with the names of the reference properties to rename.
            Default is an empty list.

    Returns:
        tuple: A tuple of two infinite iterators, one for training batches and one for validation batches.
            For each element in the batch, we expect a "species" key with the atomic numbers of the atoms in the system. Arrays are concatenated along the first axis and the following keys are added to distinguish between the systems:
            - 'natoms': np.ndarray. Array with the number of atoms in each system.
            - 'batch_index': np.ndarray. Array with the index of the system to which each atom
            if the keys "forces", "total_energy", "atomic_energies" or any of the elements in rename_refs are present, the keys are renamed by prepending "true_" to the key name.
    """

    assert isinstance(
        training_parameters, dict
    ), "training_parameters must be a dictionary."
    assert isinstance(
        rename_refs, dict
    ), "rename_refs must be a dictionary with the keys to rename."

    pbc_training = training_parameters.get("pbc_training", False)
    minimum_image = training_parameters.get("minimum_image", False)

    coordinates_ref_key = training_parameters.get("coordinates_ref_key", None)

    input_keys = [
        "species",
        "coordinates",
        "natoms",
        "batch_index",
        "total_charge",
        "flags",
    ]
    if pbc_training:
        input_keys += ["cells", "reciprocal_cells"]
    if atom_padding:
        input_keys += ["true_atoms", "true_sys"]
    if coordinates_ref_key is not None:
        input_keys += ["system_index", "system_sign"]

    flags = {f: None for f in add_flags}
    if minimum_image and pbc_training:
        flags["minimum_image"] = None

    additional_input_keys = set(training_parameters.get("additional_input_keys", []))
    additional_input_keys_ = set()
    for key in additional_input_keys:
        if key not in input_keys:
            additional_input_keys_.add(key)
    additional_input_keys = additional_input_keys_

    all_inputs = set(input_keys + list(additional_input_keys))

    extract_all_keys = ref_keys is None
    if ref_keys is not None:
        ref_keys = set(ref_keys)
        ref_keys_ = set()
        for key in ref_keys:
            if key not in all_inputs:
                ref_keys_.add(key)

    random_rotation = training_parameters.get("random_rotation", False)
    if random_rotation:
        assert np_rng is not None, "np_rng must be provided for adding noise."

        apply_rotation = {
            1: lambda x, r: x @ r,
            -1: lambda x, r: np.einsum("...kn,kj->...jn", x, r),
            2: lambda x, r: np.einsum("li,...lk,kj->...ij", r, x, r),
        }
        def rotate_2f(x,r):
            assert x.shape[-1]==6
            # select from 6 components (xx,yy,zz,xy,xz,yz) to form the 3x3 tensor
            indices = np.array([0,3,4,3,1,5,4,5,2])
            x=x[...,indices].reshape(*x.shape[:-1],3,3)
            x=np.einsum("li,...lk,kj->...ij", r, x, r)
            # select back the 6 components
            indices = np.array([[0,0],[1,1],[2,2],[0,1],[0,2],[1,2]])
            x=x[...,indices[:,0],indices[:,1]]
            return x
        apply_rotation[-2]=rotate_2f

        valid_rotations = tuple(apply_rotation.keys())
        rotated_keys = {
            "coordinates": 1,
            "forces": 1,
            "virial_tensor": 2,
            "stress_tensor": 2,
            "virial": 2,
            "stress": 2,
        }
        if pbc_training:
            rotated_keys["cells"] = 1
        user_rotated_keys = dict(training_parameters.get("rotated_keys", {}))
        for k, v in user_rotated_keys.items():
            assert (
                v in valid_rotations
            ), f"Invalid rotation type for key {k}. Valid values are {valid_rotations}"
            rotated_keys[k] = v

        # rotated_vector_keys = set(
        #     ["coordinates", "forces"]
        #     + list(training_parameters.get("rotated_vector_keys", []))
        # )
        # if pbc_training:
        #     rotated_vector_keys.add("cells")

        # rotated_tensor_keys = set(
        #     ["virial_tensor", "stress_tensor", "virial", "stress"]
        #     + list(training_parameters.get("rotated_tensor_keys", []))
        # )
        # assert rotated_vector_keys.isdisjoint(
        #     rotated_tensor_keys
        # ), "Rotated vector keys and rotated tensor keys must be disjoint."
        # rotated_keys = rotated_vector_keys.union(rotated_tensor_keys)

        print(
            "Applying random rotations to the following keys if present:",
            list(rotated_keys.keys()),
        )

        def apply_random_rotations(output, nbatch):
            euler_angles = np_rng.uniform(0.0, 2 * np.pi, (nbatch, 3))
            r = [
                Rotation.from_euler("xyz", euler_angles[i]).as_matrix().T
                for i in range(nbatch)
            ]
            for k, l in rotated_keys.items():
                if k in output:
                    for i in range(nbatch):
                        output[k][i] = apply_rotation[l](output[k][i], r[i])

    else:

        def apply_random_rotations(output, nbatch):
            pass

    flow_matching = training_parameters.get("flow_matching", False)
    if flow_matching:
        if ref_keys is not None:
            ref_keys.add("flow_matching_target")
            if "flow_matching_target" in ref_keys_:
                ref_keys_.remove("flow_matching_target")
        all_inputs.add("flow_matching_time")

        def add_flow_matching(output, nbatch):
            ts = np_rng.uniform(0.0, 1.0, (nbatch,))
            targets = []
            for i in range(nbatch):
                x1 = output["coordinates"][i]
                com = x1.mean(axis=0, keepdims=True)
                x1 = x1 - com
                x0 = np_rng.normal(0.0, 1.0, x1.shape)
                xt = (1 - ts[i]) * x0 + ts[i] * x1
                output["coordinates"][i] = xt
                targets.append(x1 - x0)
            output["flow_matching_target"] = targets
            output["flow_matching_time"] = [np.array(t) for t in ts]

    else:

        def add_flow_matching(output, nbatch):
            pass

    if pbc_training:
        print("Periodic boundary conditions are active.")
        length_nopbc = training_parameters.get("length_nopbc", 1000.0)

        def add_cell(d, output):
            if "cell" not in d:
                cell = np.asarray(
                    [
                        [length_nopbc, 0.0, 0.0],
                        [0.0, length_nopbc, 0.0],
                        [0.0, 0.0, length_nopbc],
                    ],
                    dtype=fprec,
                )
            else:
                cell = np.asarray(d["cell"], dtype=fprec)
            output["cells"].append(cell.reshape(1, 3, 3))

    else:

        def add_cell(d, output):
            if "cell" in d:
                print(
                    "Warning: 'cell' found in dataset but not training with pbc. Activate pbc_training to use periodic boundary conditions."
                )

    if extract_all_keys:

        def add_other_keys(d, output, atom_shift):
            for k, v in d.items():
                if k in ("cell", "total_charge"):
                    continue
                v_array = np.array(v)
                # Shift atom number if necessary
                if k.endswith("_atidx"):
                    v_array = v_array + atom_shift
                output[k].append(v_array)

    else:

        def add_other_keys(d, output, atom_shift):
            output["species"].append(np.asarray(d["species"]))
            output["coordinates"].append(np.asarray(d["coordinates"], dtype=fprec))
            for k in additional_input_keys:
                v_array = np.array(d[k])
                # Shift atom number if necessary
                if k.endswith("_atidx"):
                    v_array = v_array + atom_shift
                output[k].append(v_array)
            for k in ref_keys_:
                v_array = np.array(d[k])
                # Shift atom number if necessary
                if k.endswith("_atidx"):
                    v_array = v_array + atom_shift
                output[k].append(v_array)
                if k + "_mask" in d:
                    output[k + "_mask"].append(np.asarray(d[k + "_mask"]))

    def add_keys(d, output, atom_shift, batch_index):
        nat = d["species"].shape[0]

        output["natoms"].append(np.asarray([nat]))
        output["batch_index"].append(np.asarray([batch_index] * nat))
        if "total_charge" not in d:
            total_charge = np.asarray(0.0, dtype=fprec)
        else:
            total_charge = np.asarray(d["total_charge"], dtype=fprec)
        output["total_charge"].append(total_charge)

        add_cell(d, output)
        add_other_keys(d, output, atom_shift)

        return atom_shift + nat

    def collate_fn_(batch):
        output = defaultdict(list)
        atom_shift = 0
        batch_index = 0

        for d in batch:
            atom_shift = add_keys(d, output, atom_shift, batch_index)
            batch_index += 1

            if coordinates_ref_key is not None:
                output["system_index"].append(np.asarray([batch_index - 1]))
                output["system_sign"].append(np.asarray([1]))
                if coordinates_ref_key in d:
                    dref = {**d, "coordinates": d[coordinates_ref_key]}
                    atom_shift = add_keys(dref, output, atom_shift, batch_index)
                    output["system_index"].append(np.asarray([batch_index - 1]))
                    output["system_sign"].append(np.asarray([-1]))
                    batch_index += 1

        nbatch_ = len(output["natoms"])
        apply_random_rotations(output,nbatch_)
        add_flow_matching(output,nbatch_)

        # Stack and concatenate the arrays
        for k, v in output.items():
            if v[0].ndim == 0:
                v = np.stack(v)
            else:
                v = np.concatenate(v, axis=0)
            if np.issubdtype(v.dtype, np.floating):
                v = v.astype(fprec)
            output[k] = v

        if "cells" in output and pbc_training:
            output["reciprocal_cells"] = np.linalg.inv(output["cells"])

        # Rename necessary keys
        # for key in rename_refs:
        #     if key in output:
        #         output["true_" + key] = output.pop(key)
        for kold, knew in rename_refs.items():
            assert (
                knew not in output
            ), f"Cannot rename key {kold} to {knew}. Key {knew} already present."
            if kold in output:
                output[knew] = output.pop(kold)

        output["flags"] = flags
        return output

    collate_layers_train = [collate_fn_]
    collate_layers_valid = [collate_fn_]

    ### collate preprocessing
    # add noise to the training data
    noise_sigma = training_parameters.get("noise_sigma", None)
    if noise_sigma is not None:
        assert isinstance(noise_sigma, dict), "noise_sigma should be a dictionary"

        for sigma in noise_sigma.values():
            assert sigma >= 0, "Noise sigma should be a positive number"

        print("Adding noise to the training data:")
        for key, sigma in noise_sigma.items():
            print(f"  - {key} with sigma = {sigma}")

        assert np_rng is not None, "np_rng must be provided for adding noise."

        def collate_with_noise(batch):
            for key, sigma in noise_sigma.items():
                if key in batch and sigma > 0:
                    batch[key] += np_rng.normal(0, sigma, batch[key].shape).astype(
                        batch[key].dtype
                    )
            return batch

        collate_layers_train.append(collate_with_noise)

    if atom_padding:
        padder = AtomPadding(add_sys=training_parameters.get("padder_add_sys", 0))
        padder_state = padder.init()

        def collate_with_padding(batch):
            padder_state_up, output = padder(padder_state, batch)
            padder_state.update(padder_state_up)
            return output

        collate_layers_train.append(collate_with_padding)
        collate_layers_valid.append(collate_with_padding)

    if split_data_inputs:

        # input_keys += additional_input_keys
        # input_keys = set(input_keys)
        print("Input keys:", all_inputs)
        print("Ref keys:", ref_keys)

        def collate_split(batch):
            inputs = {}
            refs = {}
            for k, v in batch.items():
                if k in all_inputs:
                    inputs[k] = v
                if k in ref_keys:
                    refs[k] = v
                if k.endswith("_mask") and k[:-5] in ref_keys:
                    refs[k] = v
            return inputs, refs

        collate_layers_train.append(collate_split)
        collate_layers_valid.append(collate_split)

    ### apply all collate preprocessing
    if len(collate_layers_train) == 1:
        collate_fn_train = collate_layers_train[0]
    else:

        def collate_fn_train(batch):
            for layer in collate_layers_train:
                batch = layer(batch)
            return batch

    if len(collate_layers_valid) == 1:
        collate_fn_valid = collate_layers_valid[0]
    else:

        def collate_fn_valid(batch):
            for layer in collate_layers_valid:
                batch = layer(batch)
            return batch

    if not os.path.exists(dspath):
        raise ValueError(f"Dataset file '{dspath}' not found.")
    # dspath = training_parameters.get("dspath", None)
    print(f"Loading dataset from {dspath}...", end="")
    # print(f"   the following keys will be renamed if present : {rename_refs}")
    sharded_training = False
    if dspath.endswith(".db"):
        dataset = {}
        if train_val_split:
            dataset["training"] = DBDataset(dspath, table="training")
            dataset["validation"] = DBDataset(dspath, table="validation")
        else:
            dataset = DBDataset(dspath)
    elif dspath.endswith(".h5") or dspath.endswith(".hdf5"):
        dataset = {}
        if train_val_split:
            dataset["training"] = H5Dataset(dspath, table="training")
            dataset["validation"] = H5Dataset(dspath, table="validation")
        else:
            dataset = H5Dataset(dspath)
    elif dspath.endswith(".pkl") or dspath.endswith(".pickle"):
        with open(dspath, "rb") as f:
            dataset = pickle.load(f)
        if not train_val_split and isinstance(dataset, dict):
            dataset = dataset["training"]
    elif os.path.isdir(dspath):
        if train_val_split:
            dataset = {}
            with open(dspath + "/validation.pkl", "rb") as f:
                dataset["validation"] = pickle.load(f)
        else:
            dataset = None

        shard_files = sorted(glob.glob(dspath + "/training_*.pkl"))
        nshards = len(shard_files)
        if nshards == 0:
            raise ValueError("No dataset shards found.")
        elif nshards == 1:
            with open(shard_files[0], "rb") as f:
                if train_val_split:
                    dataset["training"] = pickle.load(f)
                else:
                    dataset = pickle.load(f)
        else:
            print(f"Found {nshards} dataset shards.")
            sharded_training = True

    else:
        raise ValueError(
            f"Unknown dataset format. Supported formats: '.db', '.h5', '.pkl', '.pickle'"
        )
    print(" done.")

    ### BUILD DATALOADERS
    # batch_size = training_parameters.get("batch_size", 16)
    shuffle = training_parameters.get("shuffle_dataset", True)
    if train_val_split:
        if batch_size_val is None:
            batch_size_val = batch_size
        dataloader_validation = DataLoader(
            dataset["validation"],
            batch_size=batch_size_val,
            shuffle=shuffle,
            collate_fn=collate_fn_valid,
        )

    if sharded_training:

        def iterate_sharded_dataset():
            indices = np.arange(nshards)
            if shuffle:
                assert np_rng is not None, "np_rng must be provided for shuffling."
                np_rng.shuffle(indices)
            for i in indices:
                filename = shard_files[i]
                print(f"# Loading dataset shard from {filename}...", end="")
                with open(filename, "rb") as f:
                    dataset = pickle.load(f)
                print(" done.")
                dataloader = DataLoader(
                    dataset,
                    batch_size=batch_size,
                    shuffle=shuffle,
                    collate_fn=collate_fn_train,
                )
                for batch in dataloader:
                    yield batch

        class DataLoaderSharded:
            def __iter__(self):
                return iterate_sharded_dataset()

        dataloader_training = DataLoaderSharded()
    else:
        dataloader_training = DataLoader(
            dataset["training"] if train_val_split else dataset,
            batch_size=batch_size,
            shuffle=shuffle,
            collate_fn=collate_fn_train,
        )

    if not infinite_iterator:
        if train_val_split:
            return dataloader_training, dataloader_validation
        return dataloader_training

    def next_batch_factory(dataloader):
        while True:
            for batch in dataloader:
                yield batch

    training_iterator = next_batch_factory(dataloader_training)
    if train_val_split:
        validation_iterator = next_batch_factory(dataloader_validation)
        return training_iterator, validation_iterator
    return training_iterator


def load_model(
    parameters: Dict[str, any],
    model_file: Optional[str] = None,
    rng_key: Optional[str] = None,
) -> FENNIX:
    """
    Load a FENNIX model from a file or create a new one.

    Args:
        parameters (dict): A dictionary of parameters for the model.
        model_file (str, optional): The path to a saved model file to load.

    Returns:
        FENNIX: A FENNIX model object.
    """
    print_model = parameters["training"].get("print_model", False)
    if model_file is None:
        model_file = parameters.get("model_file", None)
    if model_file is not None and os.path.exists(model_file):
        model = FENNIX.load(model_file, use_atom_padding=False)
        if print_model:
            print(model.summarize())
        print(f"Restored model from '{model_file}'.")
    else:
        assert (
            rng_key is not None
        ), "rng_key must be specified if model_file is not provided."
        model_params = parameters["model"]
        if isinstance(model_params, str):
            assert os.path.exists(
                model_params
            ), f"Model file '{model_params}' not found."
            model = FENNIX.load(model_params, use_atom_padding=False)
            print(f"Restored model from '{model_params}'.")
        else:
            model = FENNIX(**model_params, rng_key=rng_key, use_atom_padding=False)
        if print_model:
            print(model.summarize())
    return model


def copy_parameters(variables, variables_ref, params=[".*"]):
    def merge_params(full_path_, v, v_ref):
        full_path = "/".join(full_path_[1:]).lower()
        # status = (False, "")
        for path in params:
            if re.match(path.lower(), full_path):
            # if full_path.startswith(path.lower()) and len(path) > len(status[1]):
                return v_ref
        return v
        # return v_ref if status[0] else v

    flat = traverse_util.flatten_dict(variables, keep_empty_nodes=False)
    flat_ref = traverse_util.flatten_dict(variables_ref, keep_empty_nodes=False)
    return traverse_util.unflatten_dict(
        {
            k: merge_params(k, v, flat_ref[k]) if k in flat_ref else v
            for k, v in flat.items()
        }
    )


class TeeLogger(object):
    def __init__(self, name):
        self.file = io.TextIOWrapper(open(name, "wb"), write_through=True)
        self.stdout = None

    def __del__(self):
        self.close()

    def write(self, data):
        self.file.write(data)
        self.stdout.write(data)
        self.flush()

    def close(self):
        self.file.close()

    def flush(self):
        self.file.flush()

    def bind_stdout(self):
        if isinstance(sys.stdout, TeeLogger):
            raise ValueError("stdout already bound to a Tee instance.")
        if self.stdout is not None:
            raise ValueError("stdout already bound.")
        self.stdout = sys.stdout
        sys.stdout = self

    def unbind_stdout(self):
        if self.stdout is None:
            raise ValueError("stdout is not bound.")
        sys.stdout = self.stdout



import jax
import jax.numpy as jnp
import flax.linen as nn
from ...utils.spherical_harmonics import generate_spherical_harmonics, CG_SO3
from ..misc.encodings import SpeciesEncoding, RadialBasis
import dataclasses
import numpy as np
from typing import Dict, Union, Callable, ClassVar
from ...utils.activations import activation_from_str, tssr3
from ..misc.nets import FullyConnectedNet


class HIPNNEmbedding(nn.Module):
    """Hierarchically Interacting Particle Neural Network

    ### Reference
    Adapted from N. Lubbers, J. S. Smith and K. Barros, Hierarchical modeling of molecular energies using a deep neural network
    J. Chem. Phys. 148, 241715 (2018) (J. Chem. Phys. 148, 241715 (2018)) (https://doi.org/10.1063/1.5011181)

    """
    
    _graphs_properties: Dict
    dim: int = 80
    """The dimension of the embedding."""
    n_onsite: int = 3
    """The number of onsite layers per interaction layer."""
    nlayers: int = 2
    """The number of interaction layers."""
    lmax: int = 0
    """The maximum value degree of spherical harmonics."""
    n_message: int = 0
    """The number of layers for the message NN."""
    activation: Union[Callable, str] = "silu"
    """The activation function."""
    graph_key: str = "graph"
    """The key for the graph input."""
    embedding_key: str = "embedding"
    """The key for the embedding output."""
    species_encoding: dict = dataclasses.field(default_factory=dict)
    """The species encoding parameters. See `fennol.models.misc.encodings.SpeciesEncoding`."""
    radial_basis: dict = dataclasses.field(default_factory=lambda: {"dim": 20})
    """The radial basis parameters. See `fennol.models.misc.encodings.RadialBasis`."""
    keep_all_layers: bool = True
    """Whether to keep embeddings from each layer in the output."""
    graph_l_key: str = "graph"
    """The key for the graph input for the spherical harmonics."""

    FID: ClassVar[str] = "HIPNN"


    @nn.compact
    def __call__(self, inputs):
        species = inputs["species"]
        assert (
            len(species.shape) == 1
        ), "Species must be a 1D array (batches must be flattened)"

        graph = inputs[self.graph_key]
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]

        cutoff = self._graphs_properties[self.graph_key]["cutoff"]

        zi = SpeciesEncoding(**self.species_encoding, name="SpeciesEncoding")(species)

        act = (
            activation_from_str(self.activation)
            if isinstance(self.activation, str)
            else self.activation
        )

        distances = graph["distances"]
        if self.lmax > 0:
            filtered_l = "parent_graph" in self._graphs_properties[self.graph_l_key]

            correct_graph = (
                self.graph_l_key == self.graph_key
                or self._graphs_properties[self.graph_l_key]["parent_graph"]
                == self.graph_key
            )
            assert (
                correct_graph
            ), f"graph_l_key={self.graph_l_key} must be a subgraph of graph_key={self.graph_key}"

            graph_l = inputs[self.graph_l_key]
            Yij = generate_spherical_harmonics(lmax=self.lmax, normalize=False)(
                graph_l["vec"] / graph_l["distances"][:, None]
            )[:, None, :]
            if filtered_l:
                Yij = Yij[:, :, 1:]
            # reps_l = np.array([2 * l + 1 for l in range(1, self.lmax + 1)])

        switch = graph["switch"][:, None]
        if self.keep_all_layers:
            zis = []
        for layer in range(self.nlayers):
            ### interaction layer
            s = RadialBasis(
                **{
                    "basis": "gaussian_rinv",
                    "end": cutoff,
                    **self.radial_basis,
                    "name": f"RadialBasis_{layer}",
                }
            )(distances)

            zself = nn.Dense(self.dim, name=f"self_int_{layer}", use_bias=True)(zi)
            if self.n_message == 0:
                V = self.param(
                    f"V_{layer}",
                    jax.nn.initializers.glorot_normal(batch_axis=0),
                    (s.shape[-1], zi.shape[-1], self.dim),
                )
                mij = jnp.einsum("...j,...k,jkl->...l", s, zi[edge_dst], V)
            else:
                mij = FullyConnectedNet(
                    [self.dim] * (self.n_message + 1),
                    activation=act,
                    name=f"mij_{layer}",
                )(jnp.concatenate([zi[edge_dst], s], axis=-1))

            if self.lmax == 0:
                zi = act(zself.at[edge_src].add(mij * switch,mode="drop"))
            else:
                if filtered_l:
                    zself = zself.at[edge_src].add(mij * switch,mode="drop")
                    filter_indices = graph_l["filter_indices"]
                    mij = mij[filter_indices]

                mij = mij * graph_l["switch"][:, None]
                Mij = mij[:, :, None] * Yij
                Mi = jax.ops.segment_sum(Mij, graph_l["edge_src"], zi.shape[0])

                if filtered_l:
                    zint = jnp.zeros_like(zself)
                else:
                    zint = jax.lax.index_in_dim(Mi, 0, axis=-1, keepdims=False)
                    Mi = Mi[:, :, 1:]

                ts = self.param(
                    f"ts_{layer}",
                    lambda key, shape: jax.random.normal(key, shape, dtype=zint.dtype),
                    (self.lmax, 3),
                )  # .repeat(reps_l)

                # zint = zint + jnp.sum(ts[None,None,:]*Mi**2,axis=-1)
                for l in range(1, self.lmax + 1):
                    Ml = jax.lax.dynamic_slice_in_dim(
                        Mi, start_index=l**2 - 1, slice_size=2 * l + 1, axis=-1
                    )
                    zint = zint + ts[l, 2] * tssr3(
                        nn.Dense(self.dim, name=f"linear_{layer}_l{l}")(
                            jnp.sum(Ml**2, axis=-1)
                        )
                    )
                zi = act(zself + zint)

            ### onsite layers
            # zi = zi + FullyConnectedNet(
            #     [self.dim] * (self.n_onsite + 1),
            #     activation=act,
            #     use_bias=True,
            #     name=f"onsite_{layer}",
            # )(zi)
            for j in range(self.n_onsite):
                zi = zi + FullyConnectedNet(
                    (self.dim, self.dim),
                    activation=act,
                    use_bias=True,
                    name=f"onsite_{layer}_{j}",
                )(zi)
            if self.keep_all_layers:
                zis.append(zi)

        output = {
            **inputs,
            self.embedding_key: zi,
        }
        if self.keep_all_layers:
            output[self.embedding_key + "_layers"] = jnp.stack(zis, axis=1)
        return output



import pathlib
import jax
import jax.numpy as jnp
import flax.linen as nn
import numpy as np
from typing import Any, Dict, Union, Callable, Sequence, Optional, ClassVar
from ...utils import AtomicUnits as au
from ...utils.periodic_table import D3_COV_RADII,UFF_VDW_RADII


class RepulsionZBL(nn.Module):
    """Repulsion energy based on the Ziegler-Biersack-Littmark potential

    FID: REPULSION_ZBL

    ### Reference
    J. F. Ziegler & J. P. Biersack , The Stopping and Range of Ions in Matter

    """

    _graphs_properties: Dict
    graph_key: str = "graph"
    """The key for the graph input."""
    energy_key: Optional[str] = None
    """The key for the output energy."""
    trainable: bool = True
    """Whether the parameters are trainable."""
    _energy_unit: str = "Ha"
    """The energy unit of the model. **Automatically set by FENNIX**"""
    proportional_regularization: bool = True
    d: float = 0.46850/au.BOHR
    p: float = 0.23
    alphas: Sequence[float] = (3.19980, 0.94229, 0.40290, 0.20162)
    cs: Sequence[float] = (0.18175273, 0.5098655, 0.28021213, 0.0281697)
    cs_logits: Sequence[float] = (0.1130, 1.1445, 0.5459, -1.7514)

    FID: ClassVar[str] = "REPULSION_ZBL"

    @nn.compact
    def __call__(self, inputs):
        species = inputs["species"]
        graph = inputs[self.graph_key]
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]

        training = "training" in inputs.get("flags", {})

        rijs = graph["distances"] / au.BOHR

        d_ = self.d
        p_ = self.p
        assert len(self.alphas) == 4, "alphas must be a sequence of length 4"
        alphas_ = np.array(self.alphas, dtype=rijs.dtype)
        assert len(self.cs) == 4, "cs must be a sequence of length 4"
        cs_ =  np.array(self.cs, dtype=rijs.dtype)
        cs_ = 0.5 * cs_ / np.sum(cs_)
        if self.trainable:
            d = jnp.abs(
                self.param(
                    "d",
                    lambda key, d: jnp.asarray(d, dtype=rijs.dtype),
                    d_,
                )
            )
            p = jnp.abs(
                self.param(
                    "p",
                    lambda key, p: jnp.asarray(p, dtype=rijs.dtype),
                    p_,
                )
            )
            cs = 0.5 * jax.nn.softmax(
                self.param(
                    "cs",
                    lambda key, cs: jnp.asarray(cs, dtype=rijs.dtype),
                    np.array(self.cs_logits, dtype=rijs.dtype),
                )
            )
            alphas = jnp.abs(
                self.param(
                    "alphas",
                    lambda key, alphas: jnp.asarray(alphas, dtype=rijs.dtype),
                    alphas_,
                )
            )

            if training:
                if self.proportional_regularization:
                    reg = jnp.asarray(
                        ((1 - alphas/alphas_) ** 2).sum()
                        + ((1 - cs/cs_) ** 2).sum()
                        + (1 - p/p_) ** 2
                        + (1 - d/d_) ** 2
                    ).reshape(1)
                else:
                    reg = jnp.asarray(
                        ((alphas_ - alphas) ** 2).sum()
                        + ((cs_ - cs) ** 2).sum()
                        + (p_ - p) ** 2
                        + (d_ - d) ** 2
                    ).reshape(1)
        else:
            cs = jnp.asarray(cs_)
            alphas = jnp.asarray(alphas_)
            d = d_
            p = p_

        if "alch_group" in inputs:
            switch = graph["switch_raw"]
            lambda_v = inputs["alch_vlambda"]
            alch_group = inputs["alch_group"]
            alch_alpha = inputs.get("alch_alpha", 0.5)
            alch_m = inputs.get("alch_m", 2)

            mask = alch_group[edge_src] == alch_group[edge_dst]

            rijs = jnp.where(mask, rijs, (rijs**2 + alch_alpha**2 * (1 - lambda_v))**0.5)
            lambda_v = 0.5*(1-jnp.cos(jnp.pi*lambda_v))
            switch = jnp.where(
                mask,
                switch,
                (lambda_v**alch_m) * switch ,
            )
        else:
            switch = graph["switch"]

        Z = jnp.where(species > 0, species.astype(rijs.dtype), 0.0)
        Zij = Z[edge_src]*Z[edge_dst]
        Zp = Z**p / d
        x = rijs * (Zp[edge_src] + Zp[edge_dst])
        phi = (cs[None, :] * jnp.exp(-alphas[None, :] * x[:, None])).sum(axis=-1)

        ereppair = Zij * phi / rijs * switch

        erep_atomic = jax.ops.segment_sum(ereppair, edge_src, species.shape[0])

        energy_unit = au.get_multiplier(self._energy_unit)
        energy_key = self.energy_key if self.energy_key is not None else self.name
        output = {**inputs, energy_key: erep_atomic * energy_unit}
        if self.trainable and training:
            output[energy_key + "_regularization"] = reg

        return output
    
class RepulsionNLH(nn.Module):
    """ NLH pairwise repulsive potential with pair-specific coefficients up to Z=92

    FID: REPULSION_NLH

    ### Reference
    K. Nordlund, S. Lehtola, G. Hobler, Repulsive interatomic potentials calculated at three levels of theory, Phys. Rev. A 111, 032818 
    https://doi.org/10.1103/PhysRevA.111.032818
    """

    _graphs_properties: Dict
    graph_key: str = "graph"
    """The key for the graph input."""
    energy_key: Optional[str] = None
    """The key for the output energy."""
    _energy_unit: str = "Ha"
    """The energy unit of the model. **Automatically set by FENNIX**"""
    trainable: bool = False

    FID: ClassVar[str] = "REPULSION_NLH"

    @nn.compact
    def __call__(self, inputs):

        path = str(pathlib.Path(__file__).parent.resolve()) + "/nlh_coeffs.dat"
        DATA_NLH = np.loadtxt(path,usecols=np.arange(0,8)) 
        zmax = int(np.max(DATA_NLH[:, 0]))
        AB = np.zeros(((zmax+1)**2,6), dtype=np.float32)
        for i in range(DATA_NLH.shape[0]):
            z1 = int(DATA_NLH[i, 0])
            z2 = int(DATA_NLH[i, 1])
            AB[z1+zmax*z2] = DATA_NLH[i, 2:8]
            AB[z2+zmax*z1] = DATA_NLH[i, 2:8]
        AB = AB.reshape((zmax+1)**2, 3,2)

        species = inputs["species"]
        graph = inputs[self.graph_key]
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]
        rijs = graph["distances"]

        # coefficients (a1,a2,a3)
        CS = jnp.array(AB[:, :, 0], dtype=rijs.dtype)
        # exponents (b1,b2,b3)
        ALPHAS = jnp.array(AB[:, :, 1], dtype=rijs.dtype)

        if self.trainable:
            cfact = jnp.abs(self.param(
                "c_fact",
                lambda key: jnp.ones(CS.shape[1], dtype=CS.dtype),
            ))
            CS = CS * cfact[None,:]
            CS = CS / jnp.sum(CS, axis=1, keepdims=True)
            alphas_fact = jnp.abs(self.param(
                "alpha_fact",
                lambda key: jnp.ones(ALPHAS.shape[1], dtype=ALPHAS.dtype),
            ))
            ALPHAS = ALPHAS * alphas_fact[None,:]

        s12 = species[edge_src] + zmax*species[edge_dst]
        cs = CS[s12]
        alphas = ALPHAS[s12]

        if "alch_group" in inputs:
            switch = graph["switch_raw"]
            lambda_v = inputs["alch_vlambda"]
            alch_group = inputs["alch_group"]
            alch_alpha = inputs.get("alch_alpha", 0.)
            alch_m = inputs.get("alch_m", 2)

            mask = alch_group[edge_src] == alch_group[edge_dst]

            rijs = jnp.where(mask, rijs, (rijs**2 + alch_alpha**2 * (1 - lambda_v))**0.5)
            lambda_v = 0.5*(1-jnp.cos(jnp.pi*lambda_v))
            switch = jnp.where(
                mask,
                switch,
                (lambda_v**alch_m) * switch ,
            )
            # alphas = jnp.where(
            #     mask[:,None],
            #     alphas,
            #     lambda_v * alphas ,
            # )
        else:
            switch = graph["switch"]

        Z = jnp.where(species > 0, species.astype(rijs.dtype), 0.0)
        phi = (cs * jnp.exp(-alphas*rijs[:,None])).sum(axis=-1)

        ereppair = Z[edge_src]*Z[edge_dst] * phi * switch/rijs


        energy_unit = au.get_multiplier(self._energy_unit)
        erep_atomic = (energy_unit*0.5*au.BOHR)*jax.ops.segment_sum(ereppair, edge_src, species.shape[0])

        energy_key = self.energy_key if self.energy_key is not None else self.name
        output = {**inputs, energy_key: erep_atomic }

        return output



#!/usr/bin/env python3
"""Test the electric field implementation in FENNOL."""
import jax.numpy as jnp
from fennol import FENNIX
from jax.random import PRNGKey
from fennol.utils import AtomicUnits as Au

###############################
# Test using Tinker reference #
###############################

modules = {
    'energy': {
        'module_name': 'NEURAL_NET',
        'neurons': [32, 1],
        'input_key': 'coordinates',
    },
    'charges': {
        'module_name': 'CHEMICAL_CONSTANT',
        'value': {
            'O': -0.55824546,
            'H': 0.27912273
        },
    },
    'polarisability': {
    'module_name': 'CHEMICAL_CONSTANT',
        'value': {
            'O': 0.9479157,
            'H': 0.4158204
        },
    },
    'electric_field': {
        'module_name': 'ELECTRIC_FIELD',
    },
}

model_tinker = FENNIX(
    cutoff=5.0,
    rng_key=PRNGKey(0),
    modules=modules
)

species = jnp.array(
    [
        [8, 1, 1]
    ]
).reshape(-1)

coordinates_tinker = jnp.array(
    [
        [
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 2.0],
            [0.0, 2.0, 0.0]
        ]
    ]
).reshape(-1, 3)

natoms = jnp.array([3])
batch_index = jnp.array([0, 0, 0])
output_tinker = model_tinker(
        species=species,
        coordinates=coordinates_tinker,
        natoms=natoms,
        batch_index=batch_index
    )


def test_electric_field():
    """Test the electric field output of the polarisation model."""
    electric_field = jnp.array(
        [
            [0, -0.06404562, -0.06404562],
            [0, -0.02453007, -0.10356116],
            [0, -0.10356116, -0.02453007]
        ]
    ).flatten() * Au.BOHR**2

    assert jnp.allclose(  # noqa: S101
        output_tinker['electric_field'].flatten(),
        electric_field,
        atol=1e-6
    )


##############
# Water test #
##############

model_water = FENNIX(
    cutoff=5.0,
    rng_key=PRNGKey(0),
    modules={
        'energy': {
            'module_name': 'NEURAL_NET',
            'neurons': [32, 1],
            'input_key': 'coordinates',
        },
        'charges': {
            'module_name': 'CHEMICAL_CONSTANT',
            'value': {
                'O': -0.5,
                'H': 0.25
            },
        },
        'polarisability': {
        'module_name': 'CHEMICAL_CONSTANT',
            'value': {
                'O': 1.0,
                'H': 1.0
            },
        },
        'electric_field': {
            'module_name': 'ELECTRIC_FIELD',
        },
    }
)


a = -1 / jnp.sqrt(2)**3 + 3 / jnp.sqrt(2)**5
b = -3 / jnp.sqrt(2)**5
c = -1 / jnp.sqrt(2)**3

ao = 1 / 1.0
ah = 1 / 1.0


t_water = jnp.array(
    [
        [ao, 0, 0, 2, 0, 0, -1, 0, 0],
        [0, ao, 0, 0, -1, 0, 0, 2, 0],
        [0, 0, ao, 0, 0, -1, 0, 0, -1],

        [2, 0, 0, ah, 0, 0, a, b, 0],
        [0, -1, 0, 0, ah, 0, b, a, 0],
        [0, 0, -1, 0, 0, ah, 0, 0, c],

        [-1, 0, 0, a, b, 0, ah, 0, 0],
        [0, 2, 0, b, a, 0, 0, ah, 0],
        [0, 0, -1, 0, 0, c, 0, 0, ah]
    ]
)

coordinates_water = jnp.array(
    [
        [
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0]
        ]
    ]
).reshape(-1, 3)

output_water = model_water(
        species=species,
        coordinates=coordinates_water,
        natoms=natoms,
        batch_index=batch_index
    )


def test_electric_field_water():
    """Test the electric field output of the polarisation model."""
    electric_field = jnp.array(
        [
            [-0.12585367, -0.12585367, 0.],
            [-0.19055352, -0.06115383, 0.],
            [-0.06115383, -0.19055352, 0.]]
    ).flatten() * Au.BOHR**2

    assert jnp.allclose(  # noqa: S101
        output_water['electric_field'].flatten(),
        electric_field,
        atol=1e-5
    )



import jax
import jax.numpy as jnp
from functools import partial

def colvar_distance(coordinates, atom1, atom2):
    return jnp.linalg.norm(coordinates[atom1] - coordinates[atom2])

def colvar_angle(coordinates, atom1, atom2, atom3):
    v1 = coordinates[atom1] - coordinates[atom2]
    v2 = coordinates[atom3] - coordinates[atom2]
    v1 = v1 / jnp.linalg.norm(v1)
    v2 = v2 / jnp.linalg.norm(v2)
    return jnp.arccos(jnp.dot(v1, v2))

def colvar_dihedral(coordinates, atom1, atom2, atom3, atom4):
    v1 = coordinates[atom1] - coordinates[atom2]
    v2 = coordinates[atom3] - coordinates[atom2]
    v3 = coordinates[atom4] - coordinates[atom3]
    n1 = jnp.cross(v1, v2)
    n2 = jnp.cross(v2, v3)
    n1 = n1 / jnp.linalg.norm(n1)
    n2 = n2 / jnp.linalg.norm(n2)
    return jnp.arccos(jnp.dot(n1, n2))


def setup_colvars(colvars_definitions):
    colvars = {}
    for colvar_name, colvar_def in colvars_definitions.items():
        colvar_type = colvar_def.get("type", "distance")
        if colvar_type == "distance":
            atom1 = colvar_def["atom1"]
            atom2 = colvar_def["atom2"]
            colvars[colvar_name] = partial(colvar_distance, atom1=atom1, atom2=atom2)
        elif colvar_type == "angle":
            atom1 = colvar_def["atom1"]
            atom2 = colvar_def["atom2"]
            atom3 = colvar_def["atom3"]
            colvars[colvar_name] = partial(colvar_angle, atom1=atom1, atom2=atom2, atom3=atom3)
        elif colvar_type == "dihedral":
            atom1 = colvar_def["atom1"]
            atom2 = colvar_def["atom2"]
            atom3 = colvar_def["atom3"]
            atom4 = colvar_def["atom4"]
            colvars[colvar_name] = partial(colvar_dihedral, atom1=atom1, atom2=atom2, atom3=atom3, atom4=atom4)
        else:
            raise ValueError(f"Unknown colvar type {colvar_type}")
    
    return colvars


import numpy as np
import flax.linen as nn
import jax
import jax.numpy as jnp
import math
import optax
import os
from pathlib import Path
from flax.core import freeze, unfreeze

from ..models.fennix import FENNIX
from ..utils.atomic_units import AtomicUnits as au  # CM1,THZ,BOHR,MPROT
from ..utils import Counter
from ..utils.deconvolution import (
    deconvolute_spectrum,
    kernel_lorentz_pot,
    kernel_lorentz,
)

def initialize_ir_spectrum(simulation_parameters,system_data,fprec,dt,apply_kubo_fact=False):
    state = {}

    parameters = simulation_parameters.get("ir_parameters", {})
    dipole_model = parameters["dipole_model"]
    dipole_model = Path(str(dipole_model).strip())
    if not dipole_model.exists():
        raise FileNotFoundError(f"Dipole model file {dipole_model} not found")
    else:
        print(f"# Using '{dipole_model}' as dipole model.")
        dipole_model = FENNIX.load(dipole_model)

        nblist_skin = simulation_parameters.get("nblist_skin", -1.0)
        pbc_data = system_data.get("pbc", None)

        ### CONFIGURE PREPROCESSING
        preproc_state = unfreeze(dipole_model.preproc_state)
        layer_state = []
        for st in preproc_state["layers_state"]:
            stnew = unfreeze(st)
            if pbc_data is not None:
                stnew["minimum_image"] = pbc_data["minimum_image"]
            if nblist_skin > 0:
                stnew["nblist_skin"] = nblist_skin
            if "nblist_mult_size" in simulation_parameters:
                stnew["nblist_mult_size"] = simulation_parameters["nblist_mult_size"]
            if "nblist_add_neigh" in simulation_parameters:
                stnew["add_neigh"] = simulation_parameters["nblist_add_neigh"]
            layer_state.append(freeze(stnew))
        preproc_state["layers_state"] = layer_state
        dipole_model.preproc_state = freeze(preproc_state)


    Tseg = parameters.get("tseg", 1.0 / au.PS) * au.FS
    nseg = int(Tseg / dt)
    Tseg = nseg * dt
    dom = 2 * np.pi / (3 * Tseg)
    omegacut = parameters.get("omegacut", 15000.0 / au.CM1) / au.FS
    nom = int(omegacut / dom)
    omega = dom * np.arange((3 * nseg) // 2 + 1)

    assert (
        omegacut < omega[-1]
    ), f"omegacut must be smaller than {omega[-1]*au.CM1} CM-1"

    startsave = parameters.get("startsave", 1)
    counter = Counter(nseg, startsave=startsave)
    state["istep"] = 0
    state["nsample"] = 0
    state["nadapt"] = 0

    use_qvel = parameters.get("use_qvel", False)
    if use_qvel:
        state["qvel"] = jnp.zeros((3,), dtype=fprec)
    else:
        nat = system_data["nat"]
        state["musave-2"] = jnp.zeros((3,), dtype=fprec)
        state["musave-1"] = jnp.zeros((3,), dtype=fprec)
        state["qsave-2"] = jnp.zeros((nat,1), dtype=fprec)
        state["qsave-1"] = jnp.zeros((nat,1), dtype=fprec)
        state["pos_save-2"] = jnp.zeros((nat,3), dtype=fprec)
        state["pos_save-1"] = jnp.zeros((nat,3), dtype=fprec)
    state["mudot"] = jnp.zeros((nseg, 3), dtype=fprec)
    state["Cmumu"] = jnp.zeros((nom,), dtype=fprec)
    state["first"] = True

    kT = system_data["kT"]
    kubo_fact = np.ones_like(omega)
    if apply_kubo_fact:
        uu = 0.5*omega[1:]*au.FS/kT
        kubo_fact[1:] = np.tanh(uu)/uu

    do_deconvolution = parameters.get("deconvolution", False)
    if do_deconvolution:
        gamma = simulation_parameters.get("gamma", 1.0 / au.THZ) / au.FS
        niter_deconv = parameters.get("niter_deconv", 20)
        print("# Deconvolution of IR spectra with gamma=", gamma*1000,"ps-1 and niter=",niter_deconv)
    

    kelvin = system_data["temperature"]
    c=2.99792458e-2 # speed of light in cm/ps
    # mufact = 1000*4*np.pi**2/(3*kT*c*au.BOHR**2)
    mufact = 1000*418.40*332.063714*2*np.pi**2/(0.831446215*kelvin*3.*c)
    pbc_data = system_data.get("pbc", None)
    if pbc_data is not None:
        cell = pbc_data["cell"]
        volume = np.abs(np.linalg.det(cell)) #/au.BOHR**3
        mufact = mufact/volume


    @jax.jit
    def compute_spectra(state):
        mudot = state["mudot"]
        smu = jnp.fft.rfft(mudot, 3 * nseg, axis=0, norm="ortho")
        Cmumu = dt * jnp.sum(jnp.abs(smu[:nom]) ** 2, axis=-1)

        nsinv = 1.0 / state["nsample"]
        b1 = 1.0 - nsinv
        Cmumu = state["Cmumu"] * b1 + Cmumu * nsinv

        return {
            **state,
            "Cmumu": Cmumu,
        }

    def write_spectra_to_file(state):
        Cmumu_avg = np.array(state["Cmumu"])*mufact*kubo_fact[:nom]

        if do_deconvolution:
            s_out, s_rec, K_D = deconvolute_spectrum(
                Cmumu_avg,
                omega[:nom],
                gamma,
                niter_deconv,
                kernel=kernel_lorentz,
                trans=False,
                symmetrize=True,
                verbose=False,
                K_D=state.get("K_D", None),
            )
            state = {**state, "K_D": K_D}
            spectra=(s_out,Cmumu_avg,s_rec)
        else:
            spectra=(Cmumu_avg,)

        columns = np.column_stack(
            (
                omega[:nom] * (au.FS * au.CM1),
                *spectra,
            )
        )
        np.savetxt(
            f"IR_spectrum.out",
            columns,
            fmt="%12.6f",
            header="#omega Cmudot",
        )
        print("# IR spectrum written.")

        return state
    
    @jax.jit
    def save_dipole(q,vel,pos,dip,cell_reciprocal, state):

        q=q.reshape(-1,1)
        if use_qvel:
            qvel = (q*vel).sum(axis=0)
        if "first" in state:
            state = {
                **state,
                "musave-2":dip,
                "musave-1":dip,
            }
            if use_qvel:
                state["qvel"] = qvel
            else:
                state["pos_save-2"] = pos
                state["pos_save-1"] = pos
                state["qsave-2"] = q
                state["qsave-1"] = q
            del state["first"]

        istep = state["istep"]
        new_state = {**state, "istep": istep + 1}
        new_state["musave-2"] = state["musave-1"]
        new_state["musave-1"] = dip
        
        dipdot = (dip - state["musave-2"])/(2*dt)
        if use_qvel:
            qrdot = state["qvel"]
            new_state["qvel"] = qvel
        else:
            new_state["pos_save-2"] = state["pos_save-1"]
            new_state["pos_save-1"] = pos
            new_state["qsave-2"] = state["qsave-1"]
            new_state["qsave-1"] = q

            if cell_reciprocal is not None:
                cell,reciprocal_cell = cell_reciprocal
                vec = pos - state["pos_save-2"]
                shift = -jnp.round(jnp.dot(vec, reciprocal_cell))
                pos = pos + jnp.dot(shift, cell)

            qrdot = (q*pos - state["qsave-2"]*state["pos_save-2"]).sum(axis=0)/(2*dt)

        new_state["mudot"] = state["mudot"].at[istep].set(qrdot+dipdot)

        return new_state

    # @jax.jit
    # def save_dipole(mudot, state):

    #     istep = state["istep"]
    #     new_state = {**state, "istep": istep + 1}
    #     new_state["mudot"] = state["mudot"].at[istep].set(mudot)

    #     return new_state
    
    def postprocess(state):
        counter.increment()
        if not counter.is_reset_step:
            return state
        state["nadapt"] += 1
        state["nsample"] = max(state["nadapt"] - startsave + 1, 1)
        state = compute_spectra(state)
        state["istep"] = 0
        state = write_spectra_to_file(state)
        return state
    
    return dipole_model,state,save_dipole, postprocess
        
        


import jax
import jax.numpy as jnp
import flax.linen as nn
from ...utils.spherical_harmonics import generate_spherical_harmonics
from ..misc.encodings import SpeciesEncoding, RadialBasis
import dataclasses
import numpy as np
from typing import Any, Dict, List, Union, Callable, Tuple, Sequence, Optional, ClassVar
from ..misc.nets import FullyConnectedNet
from ..misc.e3 import (
    FilteredTensorProduct,
    ChannelMixingE3,
    ChannelMixing,
    E3NN_AVAILABLE,
    E3NN_EXCEPTION,
    Irreps,
)


class AllegroEmbedding(nn.Module):
    """Allegro equivariant pair embedding

    FID : ALLEGRO

    ### Reference
    Musaelian, A., Batzner, S., Johansson, A. et al. Learning local equivariant representations for large-scale atomistic dynamics.
      Nat Commun 14, 579 (2023). https://doi.org/10.1038/s41467-023-36329-y

    """

    _graphs_properties: Dict
    dim: int = 128
    """ The dimension of the embedding."""
    nchannels: int = 16
    """ The number of equivariant channels."""
    nlayers: int = 3
    """ The number of interaction layers."""
    lmax: int = 2
    """ The maximum degree of tensorial embedding."""
    lmax_density: Optional[int] = None
    """ The maximum degree of spherical harmonics for density.
        If None, it will be set to lmax. Must be greater or equal to lmax."""
    twobody_hidden: Sequence[int] = dataclasses.field(default_factory=lambda: [128])
    """ The number of hidden neurons in the two-body network."""
    embedding_hidden: Sequence[int] = dataclasses.field(default_factory=lambda: [])
    """ The number of hidden neurons in the embedding network."""
    latent_hidden: Sequence[int] = dataclasses.field(default_factory=lambda: [128])
    """ The number of hidden neurons in the latent network."""
    activation: Union[Callable, str] = "silu"
    """ The activation function to use."""
    graph_key: str = "graph"
    """ The key in the input dictionary that corresponds to the graph."""
    embedding_key: str = "embedding"
    """ The key to use for the output embedding in the returned dictionary."""
    tensor_embedding_key: str = "tensor_embedding"
    """ The key to use for the output tensor embedding in the returned dictionary."""
    species_encoding: dict = dataclasses.field(default_factory=dict)
    """ The species encoding parameters.  See `fennol.models.misc.encodings.SpeciesEncoding`"""
    radial_basis: dict = dataclasses.field(default_factory=dict)
    """ The radial basis parameters. See `fennol.models.misc.encodings.RadialBasis`"""

    FID: ClassVar[str] = "ALLEGRO"

    @nn.compact
    def __call__(self, inputs):
        """ Forward pass of the Allegro model. """
        species = inputs["species"]

        graph = inputs[self.graph_key]
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]
        switch = graph["switch"][:, None]
        cutoff = self._graphs_properties[self.graph_key]["cutoff"]
        radial_basis = RadialBasis(
            **{**self.radial_basis, "end": cutoff, "name": "RadialBasis"}
        )(graph["distances"])

        species_encoding = SpeciesEncoding(
            **self.species_encoding, name="SpeciesEncoding"
        )(species)

        xij = (
            FullyConnectedNet(
                neurons=[*self.twobody_hidden, self.dim], activation=self.activation
            )(
                jnp.concatenate(
                    [
                        species_encoding[edge_src],
                        species_encoding[edge_dst],
                        radial_basis,
                    ],
                    axis=-1,
                )
            )
            * switch
        )

        lmax_density = self.lmax_density if self.lmax_density is not None else self.lmax
        assert lmax_density >= self.lmax

        Yij = generate_spherical_harmonics(lmax=lmax_density, normalize=False)(
            graph["vec"] / graph["distances"][:, None]  
        )[:, None, :]

        nel = (self.lmax + 1) ** 2
        Vij = (
            ChannelMixingE3(self.lmax, 1, self.nchannels)(Yij[..., :nel])
            * nn.Dense(self.nchannels, use_bias=False)(xij)[:, :, None]
        )

        for _ in range(self.nlayers):
            rhoij = (
                FullyConnectedNet(
                    neurons=[*self.embedding_hidden, self.nchannels],
                    activation=self.activation,
                )(xij)
                * switch
            )[:, :, None] * Yij
            density = (
                jnp.zeros((species.shape[0], *rhoij.shape[1:])).at[edge_src].add(rhoij)
            )

            Lij = FilteredTensorProduct(self.lmax, lmax_density)(Vij, density[edge_src])
            scals = jax.lax.index_in_dim(Lij, 0, axis=-1, keepdims=False)
            lij = FullyConnectedNet(neurons=[*self.latent_hidden, self.dim])(
                jnp.concatenate((xij, scals), axis=-1)
            )

            xij = xij + lij * switch
            Vij = ChannelMixing(self.lmax, self.nchannels, self.nchannels)(Lij)

        if self.embedding_key is None:
            return xij, Vij
        return {**inputs, self.embedding_key: xij, self.tensor_embedding_key: Vij}


if E3NN_AVAILABLE:
    import e3nn_jax as e3nn

    class AllegroE3NNEmbedding(nn.Module):
        """Allegro equivariant pair embedding

        FID : ALLEGRO_E3NN

        in this version, equivariant operations use the e3nn library.

        Reference
        ---------
        Musaelian, A., Batzner, S., Johansson, A. et al. Learning local equivariant representations for large-scale atomistic dynamics.
        Nat Commun 14, 579 (2023). https://doi.org/10.1038/s41467-023-36329-y

        """

        _graphs_properties: Dict
        dim: int = 128
        """ The dimension of the embedding."""
        nchannels: int = 16
        """ The number of equivariant channels."""
        nlayers: int = 3
        """ The number of interaction layers."""
        irreps_Vij: Union[str, int, 'Irreps'] = 2
        """ Irreps used for the tensor embedding."""
        lmax_density: int = None
        """ The maximum degree of spherical harmonics for density.
            If None, it will be set to lmax. Must be greater or equal to lmax."""
        twobody_hidden: Sequence[int] = dataclasses.field(default_factory=lambda: [128])
        """ The number of hidden neurons in the two-body network."""
        embedding_hidden: Sequence[int] = dataclasses.field(default_factory=lambda: [])
        """ The number of hidden neurons in the embedding network."""
        latent_hidden: Sequence[int] = dataclasses.field(default_factory=lambda: [128])
        """ The number of hidden neurons in the latent network."""
        activation: Union[Callable, str] = "silu"
        """ The activation function to use."""
        graph_key: str = "graph"
        """ The key in the input dictionary that corresponds to the graph."""
        embedding_key: str = "embedding"
        """ The key to use for the output embedding in the returned dictionary."""
        tensor_embedding_key: str = "tensor_embedding"
        """ The key to use for the output tensor embedding in the returned dictionary."""
        species_encoding: dict = dataclasses.field(default_factory=dict)
        """ The species encoding parameters."""
        radial_basis: dict = dataclasses.field(default_factory=dict)
        """ The radial basis parameters."""

        FID: ClassVar[str] = "ALLEGRO_E3NN"
        """ Identification of the module when building a model."""

        @nn.compact
        def __call__(self, inputs):
            species = inputs["species"]
            assert (
                len(species.shape) == 1
            ), "Species must be a 1D array (batches must be flattened)"

            graph = inputs[self.graph_key]
            edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]
            switch = graph["switch"][:, None]
            cutoff = self._graphs_properties[self.graph_key]["cutoff"]
            radial_basis = RadialBasis(
                **{**self.radial_basis, "end": cutoff, "name": "RadialBasis"}
            )(graph["distances"])
            radial_size = radial_basis.shape[-1]

            species_encoding = SpeciesEncoding(
                **self.species_encoding, name="SpeciesEncoding"
            )(species)
            afvs_size = species_encoding.shape[-1]

            xij = (
                FullyConnectedNet(
                    neurons=[*self.twobody_hidden, self.dim], activation=self.activation
                )(
                    jnp.concatenate(
                        [
                            species_encoding[edge_src],
                            species_encoding[edge_dst],
                            radial_basis,
                        ],
                        axis=-1,
                    )
                )
                * switch
            )
            if isinstance(self.irreps_Vij, int):
                irreps_Vij = e3nn.Irreps.spherical_harmonics(self.irreps_Vij)
            elif isinstance(self.irreps_Vij, str):
                irreps_Vij = e3nn.Irreps(self.irreps_Vij)
            else:
                irreps_Vij = self.irreps_Vij
            lmax = max(irreps_Vij.ls)
            lmax_density = self.lmax_density or lmax
            irreps_density = e3nn.Irreps.spherical_harmonics(lmax_density)

            # Yij = e3nn.IrrepsArray(
            #     irreps_density,
            #     generate_spherical_harmonics(lmax=lmax_density, normalize=False)(
            #         graph["vec"] / graph["distances"][:, None]
            #     ),
            # )[:, None, :]
            Yij = e3nn.spherical_harmonics(
                irreps_density, graph["vec"], normalize=True
            )[:, None, :]

            Vij = (
                e3nn.flax.Linear(irreps_Vij, channel_out=self.nchannels)(Yij)
                * nn.Dense(self.nchannels, use_bias=False)(xij)[:, :, None]
            )

            for _ in range(self.nlayers):
                rhoij = (
                    FullyConnectedNet(
                        neurons=[*self.embedding_hidden, self.nchannels],
                        activation=self.activation,
                    )(xij)
                    * switch
                )[:, :, None] * Yij
                density = e3nn.scatter_sum(
                    rhoij, dst=edge_src, output_size=species_encoding.shape[0]
                )

                Lij = e3nn.tensor_product(
                    Vij, density[edge_src], filter_ir_out=irreps_Vij
                )
                scals = Lij.filter(["0e"]).array.reshape(Lij.shape[0], -1)
                lij = FullyConnectedNet(neurons=[*self.latent_hidden, self.dim])(
                    jnp.concatenate((xij, scals), axis=-1)
                )

                xij = xij + lij * switch
                # filtering
                Lij = e3nn.flax.Linear(irreps_Vij)(Lij)
                # channel mixing
                Vij = e3nn.flax.Linear(irreps_Vij, channel_out=self.nchannels)(Lij)

            if self.embedding_key is None:
                return xij, Vij
            return {**inputs, self.embedding_key: xij, self.tensor_embedding_key: Vij}

else:

    class AllegroE3NNEmbedding(nn.Module):
        FID: ClassVar[str] = "ALLEGRO_E3NN"

        def __call__(self, *args, **kwargs) -> Any:
            raise E3NN_EXCEPTION



import jax
import jax.numpy as jnp
import flax.linen as nn
import numpy as np
from typing import Any, Dict, Union, Callable, Sequence, Optional, ClassVar
from ...utils import AtomicUnits as au
from ...utils.periodic_table import (
    D3_ELECTRONEGATIVITIES,
    D3_HARDNESSES,
    D3_VDW_RADII,
    D3_COV_RADII,
    D3_KAPPA,
    VDW_RADII,
    POLARIZABILITIES,
    C6_FREE,
    VALENCE_ELECTRONS,
)


class VdwOQDO(nn.Module):
    """ Dispersion and exchange based on the Optimized Quantum Drude Oscillator model.
    
    FID : VDW_OQDO

    ### Reference
    A. Khabibrakhmanov, D. V. Fedorov, and A. Tkatchenko, Universal Pairwise Interatomic van der Waals Potentials Based on Quantum Drude Oscillators,
    J. Chem. Theory Comput. 2023, 19, 21, 7895–7907 (https://doi.org/10.1021/acs.jctc.3c00797)
    
    """
    graph_key: str = "graph"
    """ The key for the graph input."""
    include_exchange: bool = True
    """ Whether to compute the exchange part."""
    ratiovol_key: Optional[str] = None
    """ The key for the ratio between AIM volume and free-atom volume. 
         If None, the volume ratio is assumed to be 1.0."""
    energy_key: Optional[str] = None
    """ The key for the output energy. If None, the name of the module is used."""
    damped: bool = True
    """ Whether to use short-range damping."""
    _energy_unit: str = "Ha"
    """The energy unit of the model. **Automatically set by FENNIX**"""

    FID: ClassVar[str]  = "VDW_OQDO"

    @nn.compact
    def __call__(self, inputs):
        energy_unit = au.get_multiplier(self._energy_unit)

        species = inputs["species"]
        graph = inputs[self.graph_key]
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]
        rij = graph["distances"] / au.BOHR
        switch = graph["switch"]

        c6 = jnp.asarray(C6_FREE)[species]
        alpha = jnp.asarray(POLARIZABILITIES)[species]

        if self.ratiovol_key is not None:
            ratiovol = inputs[self.ratiovol_key] + 1.0e-6
            if ratiovol.shape[-1] == 1:
                ratiovol = jnp.squeeze(ratiovol, axis=-1)
            c6 = c6 * ratiovol**2
            alpha = alpha * ratiovol

        c6i, c6j = c6[edge_src], c6[edge_dst]
        alphai, alphaj = alpha[edge_src], alpha[edge_dst]

        # combination rules
        alphaij = 0.5 * (alphai + alphaj)
        c6ij = 2 * alphai * alphaj * c6i * c6j / (c6i * alphaj**2 + c6j * alphai**2)

        # equilibrium distance
        Re = (alphaij * (128.0 / au.FSC ** (4.0 / 3.0))) ** (1.0 / 7.0)
        Re2 = Re**2
        Re4 = Re**4
        # fit to largest root of eq (S33) of "Universal Pairwise Interatomic van der Waals Potentials Based On Quantum Drude Oscillators"
        if self.damped:
            muw = (
                4.83053463e-01
                - 3.76191669e-02 * Re
                + 1.27066988e-03 * Re2
                - 7.21940151e-07 * Re4
            ) / (3.84212120e-02 - 3.16915319e-02 * Re + 2.37410890e-02 * Re2)
        else:
            muw = (
                3.66316787e01
                - 5.79579187 * Re
                + 3.02674813e-01 * Re2
                - 3.65461255e-04 * Re4
            ) / (-1.46169102e01 + 7.32461225 * Re)

        c8ij = 5 * c6ij / muw
        c10ij = 245 * c6ij / (8 * muw**2)

        if self.damped:
            z = 0.5 * muw * rij**2
            ez = jnp.exp(-z)
            f6 = 1.0 - ez * (1.0 + z + 0.5 * z**2 + (1.0 / 6.0) * z**3)
            f8 = f6 - (1.0 / 24.0) * ez * z**4
            f10 = f8 - (1.0 / 120.0) * ez * z**5
            epair = (
                f6 * c6ij / rij**6 + f8 * c8ij / rij**8 + f10 * c10ij / rij**10
            )
        else:
            epair = c6ij / rij**6 + c8ij / rij**8 + c10ij / rij**10

        edisp = (-0.5*energy_unit) * jax.ops.segment_sum(epair * switch, edge_src, species.shape[0])

        output_key = self.name if self.energy_key is None else self.energy_key

        if not self.include_exchange:
            return {**inputs, output_key: edisp}

        ### exchange
        w = 4 * c6ij / (3 * alphaij**2)
        # q = (alphaij * mu*w**2)**0.5
        q2 = alphaij * muw * w
        # undamped case
        if self.damped:
            ze = 0.5 * muw * Re2
            eze = jnp.exp(-ze)

            s6 = eze * (1.0 + ze + 0.5 * ze**2 + (1.0 / 6.0) * ze**3)
            f6e = 1.0 - s6
            muwRe = muw * Re
            df6e = muwRe * s6 - eze * (
                muwRe + 0.5 * Re * muwRe**2 + (1.0 / 8.0) * Re2 * muwRe**3
            )

            s8 = (1.0 / 24.0) * eze * ze**4
            f8e = f6e - s8
            df8e = df6e + muwRe * s8 - (1.0 / 48.0) * eze * Re2 * Re * muwRe**4

            s10 = (1.0 / 120.0) * eze * ze**5
            f10e = f8e - s10
            df10e = df8e + muwRe * s10 - (1.0 / 384.0) * eze * Re2 * Re2 * muwRe**5

            den = 2 * c6ij * Re2 * (6 * f6e - Re * df6e)
            A = (
                0.5
                + c8ij * (8 * f8e - Re * df8e) / den
                + c10ij * (10 * f10e - Re * df10e) / (den * Re2)
            )
        else:
            A = 0.5 + 2 * c8ij / (3 * c6ij * Re2) + 5 * c10ij / (6 * c6ij * Re4)
            ez = jnp.exp(-0.5 * muw * rij**2)

        exij = A * q2 * ez / rij
        ex = (0.5*energy_unit) * jax.ops.segment_sum(exij * switch, edge_src, species.shape[0])

        return {
            **inputs,
            output_key + "_dispersion": edisp,
            output_key + "_exchange": ex,
            output_key: edisp + ex,
        }



import time
import numpy as np


def xyz_reader(
    filename,
    has_comment_line=False,
    indexed=False,
    start=1,
    stop=-1,
    step=1,
    max_frames=None,
    stream=False,
    interval=2.0,
    sleep=time.sleep,
):
    if stop > 0 and start > stop:
        return
    pf = open(filename, "r")
    inside_frame = False
    nat_read = False
    iframe = 0
    stride = step
    nframes = 0
    if start > 1:
        print(f"skipping {start-1} frames")
    if not has_comment_line:
        comment_line = ""
    while True:
        line = pf.readline()

        if not line:
            if stream:
                sleep(interval)
                continue
            elif inside_frame or nat_read:
                raise Exception("Error: premature end of file!")
            else:
                break

        line = line.strip()

        if not inside_frame:
            if not nat_read:
                if line.startswith("#") or line == "":
                    continue
                nat = int(line.split()[0])
                nat_read = True
                iat = 0
                inside_frame = not has_comment_line
                xyz = np.zeros((nat, 3))
                symbols = []
                continue

            # box = np.array(line.split(), dtype="float")
            comment_line = line
            inside_frame = True
            continue

        if line.startswith("#") or line == "":
            raise Exception("Error: premature end of frame!")

        if not nat_read:
            raise Exception("Error: nat not read!")
        ls = line.split()
        iat, s = (int(ls[0]), 1) if indexed else (iat + 1, 0)
        symbols.append(ls[s])
        xyz[iat - 1, :] = np.array([ls[s + 1], ls[s + 2], ls[s + 3]], dtype="float")
        if iat == nat:
            iframe += 1
            inside_frame = False
            nat_read = False
            if iframe < start:
                continue
            stride += 1
            if stride >= step:
                nframes += 1
                stride = 0
                yield symbols, xyz, comment_line
            if (max_frames is not None and nframes >= max_frames) or (
                stop > 0 and iframe >= stop
            ):
                break


def read_xyz(
    filename,
    has_comment_line=False,
    indexed=False,
    start=1,
    stop=-1,
    step=1,
    max_frames=None,
):
    return [
        frame
        for frame in xyz_reader(
            filename, has_comment_line, indexed, start, stop, step, max_frames
        )
    ]


def last_xyz_frame(filename, has_comment_line=False, indexed=False):
    last_frame = None
    for frame in xyz_reader(
        filename, has_comment_line, indexed, start=-1, stop=-1, step=1, max_frames=1
    ):
        last_frame = frame
    return last_frame


def write_arc_frame(
    f,
    symbols,
    coordinates,
    types=None,
    nbonds=None,
    connectivity=None,
    cell=None,
    **kwargs,
):
    nat = len(symbols)
    f.write(f"{nat}\n")
    if cell is not None:
        f.write(" ".join([f"{x: 15.5f}" for x in cell.flatten()]) + "\n")
    # f.write(f'{axis} {axis} {axis} 90.0 90.0 90.0 \n')
    for i in range(nat):
        line = f"{i+1} {symbols[i]:3} {coordinates[i,0]: 15.3f} {coordinates[i,1]: 15.3f} {coordinates[i,2]: 15.3f}"
        if types is not None:
            line += f"   {types[i]}"
        if connectivity is not None and nbonds is not None:
            line += "  " + " ".join([str(x + 1) for x in connectivity[i, : nbonds[i]]])
        f.write(line + "\n")
    f.flush()


def write_extxyz_frame(
    f, symbols, coordinates, cell=None, properties={}, forces=None, **kwargs
):
    nat = len(symbols)
    f.write(f"{nat}\n")
    comment_line = ""
    if cell is not None:
        comment_line += (
            'Lattice="' + " ".join([f"{x:.3f}" for x in cell.flatten()]) + '" '
        )
    comment_line += "Properties=species:S:1:pos:R:3"
    if forces is not None:
        comment_line += ":forces:R:3"
    comment_line += " "
    for k, v in properties.items():
        comment_line += f"{k}={v} "
    f.write(f"{comment_line}\n")
    for i in range(nat):
        line = f"{symbols[i]:3} {coordinates[i,0]: 15.5e} {coordinates[i,1]: 15.5e} {coordinates[i,2]: 15.5e}"
        if forces is not None:
            line += f" {forces[i,0]: 15.5e} {forces[i,1]: 15.5e} {forces[i,2]: 15.5e}"
        f.write(f"{line}\n")
    f.flush()


def write_xyz_frame(f, symbols, coordinates,cell=None, **kwargs):
    nat = len(symbols)
    f.write(f"{nat}\n")
    if cell is not None:
        f.write(" ".join([f"{x:.3f}" for x in cell.flatten()]))
    f.write("\n")
    for i in range(nat):
        f.write(
            f"{symbols[i]:3} {coordinates[i,0]: 15.5e} {coordinates[i,1]: 15.5e} {coordinates[i,2]: 15.5e}\n"
        )
    f.flush()


def human_time_duration(seconds: float):
    """Convert seconds (duration) to human readable string

    from https://gist.github.com/borgstrom/936ca741e885a1438c374824efb038b3
    """

    if seconds < 1.0:
        return f"{seconds*1000:.3g} ms"
    if seconds < 10.0:
        return f"{seconds:.3g} s"

    TIME_DURATION_UNITS = (
        ("week", "s", 60 * 60 * 24 * 7),
        ("day", "s", 60 * 60 * 24),
        ("h", "", 60 * 60),
        ("min", "", 60),
        ("s", "", 1),
    )
    parts = []
    for unit, plur, div in TIME_DURATION_UNITS:
        amount, seconds = divmod(int(seconds), div)
        if amount > 0:
            parts.append(f"{amount} {unit}{plur if amount > 1 else ''}")
    return " ".join(parts)



import jax
import jax.numpy as jnp
import flax.linen as nn
from ...utils.spherical_harmonics import generate_spherical_harmonics
from ..misc.encodings import SpeciesEncoding, RadialBasis
import dataclasses
import numpy as np
from typing import Any, Dict, List, Union, Callable, Tuple, Sequence, Optional, ClassVar
from ..misc.nets import FullyConnectedNet
from ..misc.e3 import FilteredTensorProduct, ChannelMixingE3


class CaimanEmbedding(nn.Module):
    """Covariant Atom-In-Molecule Network

    FID : CAIMAN

    This is an E(3) equivariant embedding that forms an equivariant neighbor density
    and then uses multiple self-interaction tensor products to generate a tensorial embedding
    along with a scalar embedding (similar to the tensor/scalar tracks in allegro).

    """

    _graphs_properties: Dict
    dim: int = 128
    """ The dimension of the embedding. """
    nchannels: int = 16
    """ The number of channels. """
    nchannels_density: Optional[int] = None
    """ The number of channels for the neighborhood density. If None, it is equal to nchannels."""
    nlayers: int = 3
    """ The number of layers. """
    lmax: int = 2
    """ The maximum order of spherical tensors. """
    embedding_hidden: Sequence[int] = dataclasses.field(default_factory=lambda: [])
    """ The hidden layers for the embedding."""
    latent_hidden: Sequence[int] = dataclasses.field(default_factory=lambda: [128])
    """ The hidden layers for the latent network."""
    activation: Union[Callable, str] = "silu"
    """ The activation function."""
    graph_key: str = "graph"
    """ The key for the graph input."""
    embedding_key: str = "embedding"
    """ The key for the embedding output."""
    tensor_embedding_key: str = "tensor_embedding"
    """ The key for the tensor embedding output."""
    species_encoding: dict = dataclasses.field(default_factory=dict)
    """ The species encoding parameters. See `fennol.models.misc.encodings.SpeciesEncoding`"""
    radial_basis: dict = dataclasses.field(default_factory=dict)
    """ The radial basis parameters. See `fennol.models.misc.encodings.RadialBasis`"""
    message_passing: bool = False
    """ Whether to use message passing."""

    FID: ClassVar[str] = "CAIMAN"

    @nn.compact
    def __call__(self, inputs):
        species = inputs["species"]
        assert (
            len(species.shape) == 1
        ), "Species must be a 1D array (batches must be flattened)"
        nchannels_density = (
            self.nchannels_density
            if self.nchannels_density is not None
            else self.nchannels
        )

        graph = inputs[self.graph_key]
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]
        switch = graph["switch"][:, None]
        cutoff = self._graphs_properties[self.graph_key]["cutoff"]
        radial_basis = RadialBasis(
            **{**self.radial_basis, "end": cutoff, "name": "RadialBasis"}
        )(graph["distances"])

        Dij = (
            nn.Dense(nchannels_density, use_bias=True, name="Dij")(radial_basis)
            * switch
        )

        species_encoding = SpeciesEncoding(
            **self.species_encoding, name="SpeciesEncoding"
        )(species)

        xi = FullyConnectedNet(
            neurons=[*self.embedding_hidden, self.dim],
            activation=self.activation,
            use_bias=True,
            name="species_embedding",
        )(species_encoding)
        Zs, Zd = jnp.split(
            nn.Dense(2 * nchannels_density, use_bias=True, name="species_linear")(xi),
            2,
            axis=-1,
        )
        xij = Zs[edge_src] * Zd[edge_dst] * Dij

        Yij = generate_spherical_harmonics(lmax=self.lmax, normalize=False)(
            graph["vec"] / graph["distances"][:, None]
        )[:, None, :]

        rhoij = xij[:, :, None] * Yij

        nrep = np.array([2 * l + 1 for l in range(self.lmax + 1)])
        wsh = self.param(
            "wsh",
            lambda key, shape: jax.random.normal(key, shape, dtype=jnp.float32),
            (nchannels_density, self.lmax + 1),
        ).repeat(nrep, axis=-1)
        density = (
            jax.ops.segment_sum(rhoij, edge_src, species.shape[0]) * wsh[None, :, :]
        )

        nel = (self.lmax + 1) ** 2
        Vi = ChannelMixingE3(self.lmax, nchannels_density, self.nchannels)(
            density[..., :nel]
        )
        lambda_message = self.param(
            "lambda_message",
            lambda key: jnp.asarray(0.1, dtype=density.dtype),
        )

        for layer in range(self.nlayers):
            if self.message_passing:
                Zs, Zs = jnp.split(
                    nn.Dense(
                        2 * nchannels_density,
                        use_bias=True,
                        name=f"message_linear_{layer}",
                    )(xi),
                    2,
                    axis=-1,
                )
                mij = (
                    nn.Dense(
                        nchannels_density, use_bias=False, name=f"radial_linear_{layer}"
                    )(Dij)
                    * Zs[edge_src]
                    * Zd[edge_dst]
                )
                rhoij = (
                    mij[:, :, None]
                    * ChannelMixingE3(
                        self.lmax,
                        self.nchannels,
                        nchannels_density,
                        name=f"message_mixing_{layer}",
                    )(Vi)[edge_dst]
                )
                rhoi = jax.ops.segment_sum(rhoij, edge_src, species.shape[0])
                density = density + lambda_message * ChannelMixingE3(
                    self.lmax,
                    nchannels_density,
                    nchannels_density,
                    name=f"density_update_{layer}",
                )(rhoi)

            Hi = ChannelMixingE3(
                self.lmax,
                nchannels_density,
                self.nchannels,
                name=f"density_mixing_{layer}",
            )(density)

            Li = FilteredTensorProduct(
                self.lmax, self.lmax, self.lmax, name=f"TP_{layer}"
            )(Vi, Hi)
            scals = jax.lax.index_in_dim(Li, 0, axis=-1, keepdims=False)
            li = FullyConnectedNet(
                [*self.latent_hidden, self.dim],
                activation=self.activation,
                use_bias=True,
                name=f"latent_net_{layer}",
            )(jnp.concatenate((xi, scals), axis=-1))

            xi = xi + li
            Vi = Vi + ChannelMixingE3(
                self.lmax, self.nchannels, self.nchannels, name=f"mixing_{layer}"
            )(Li)

        if self.embedding_key is None:
            return xi, Vi
        return {**inputs, self.embedding_key: xi, self.tensor_embedding_key: Vi}
